#include "GameLoop.h"
#include "fmod/FmodManager.h"
#include "TeetairApplication.h"

#include <osg/FrameStamp>

void GameLoop::start(TeetairApplication* application) {
	instance()->app = application;
	instance()->timer.start(1000.f / application->fps());
}

void GameLoop::lock() {
	instance()->frameMutex.lock();
}

void GameLoop::unlock() {
	instance()->frameMutex.unlock();
}

FrameStamp* GameLoop::getFrameStamp() {
	return instance()->app->getFrameStamp();
}

GameLoop* GameLoop::instance() {
	static GameLoop* instance = new GameLoop();
	return instance;
}

GameLoop::GameLoop()
: app(nullptr)
, frameMutex(QMutex::Recursive)
, timeListeners()
, timeListenerInvocations()
, frameListeners()
, deferredListenerChanges()
, deferredListenerChangesMutex()
, paused(false)
{
	connect(&timer, &QTimer::timeout, this, &GameLoop::loop);
}

void GameLoop::loop() {
	if (app == nullptr)
		return;
	frameMutex.lock();
	FmodManager::system()->update();

	deferredListenerChangesMutex.lock();
	while (!deferredListenerChanges.empty()) {
		deferredListenerChanges.front()();
		deferredListenerChanges.pop();
	}
	deferredListenerChangesMutex.unlock();

	if (!paused) {
		fireFrameListenerEvents();
		fireTimeListenerEvents();
		app->frameEvent();
		app->simulate(app->simulationStep());
		app->updatePoses();
	}

	app->getViewerBase()->frame();
	frameMutex.unlock();
}

void GameLoop::fireTimeListenerEvents() {
	for (std::unordered_map<double,int>::iterator i = timeListenerInvocations.begin(); i != timeListenerInvocations.end(); ++i) {
		int invocations = getFrameStamp()->getSimulationTime() / i->first;
		if (invocations > i->second) {
			timeListenerInvocations[i->first] = invocations;
			fireTimeListenerEvents(i->first);
		}
	}
}

void GameLoop::fireTimeListenerEvents(double stepsize) {
	for (QMultiMap<double, std::function<void()>*>::iterator i = timeListeners.find(stepsize); i != timeListeners.end() && i.key() == stepsize; ++i) {
		(*i.value())();
	}
}

void GameLoop::registerForTimedEvents(std::function<void()>& eventHandler, double stepsize) {
	instance()->deferredListenerChangesMutex.lock();
	std::function<void()>* handler = &eventHandler;
	instance()->deferredListenerChanges.push(
		[handler, stepsize](){
		instance()->timeListeners.insert(stepsize, handler);
		instance()->timeListenerInvocations[stepsize] = getFrameStamp()->getSimulationTime() / stepsize;
	});
	instance()->deferredListenerChangesMutex.unlock();
}

void GameLoop::unregisterFromTimedEvents(std::function<void()>& eventHandler, double stepsize) {
	std::function<void()>* handler = &eventHandler;
	instance()->deferredListenerChangesMutex.lock();
	instance()->deferredListenerChanges.push(
		[handler, stepsize](){
		instance()->timeListeners.remove(stepsize, handler);
	});
	instance()->deferredListenerChangesMutex.unlock();
}

void GameLoop::fireFrameListenerEvents() {
	for (std::unordered_set<std::function<void()>*>::iterator i = frameListeners.begin(); i != frameListeners.end(); ++i) {
		(**i)();
	}
}

void GameLoop::registerForFrameEvents(std::function<void()>& eventHandler) {
	std::function<void()>* handler = &eventHandler;
	instance()->deferredListenerChangesMutex.lock();
	instance()->deferredListenerChanges.push(
		[handler](){
		instance()->frameListeners.insert(handler);
	});
	instance()->deferredListenerChangesMutex.unlock();
}

void GameLoop::unregisterFromFrameEvents(std::function<void()>& eventHandler) {
	std::function<void()>* handler = &eventHandler;
	instance()->deferredListenerChangesMutex.lock();
	instance()->deferredListenerChanges.push(
		[handler](){
		instance()->frameListeners.erase(handler);
	});
	instance()->deferredListenerChangesMutex.unlock();
}

void GameLoop::pause() {
	instance()->paused = true;
}

void GameLoop::unpause() {
	instance()->paused = false;
}