#pragma once

#include <QObject>
#include <QTimer>
#include <QMutex>
#include <QMultiMap>

#include <osg/FrameStamp>

#include <functional>
#include <unordered_map>
#include <unordered_set>
#include <queue>

class TeetairApplication;

/// <summary>
/// The Singleton GameLoop provides events for the running application, including the main loop for scene updates and rendering as well as time- and framebased events for registered handlers. Furthermore it supports pausing and unpausing of the simulation.
/// </summary>
class GameLoop : public QObject{
	Q_OBJECT

public:
	/// <summary>
	/// Start the game loop for the given TeetairApplication.
	/// </summary>
	static void start(TeetairApplication* application);
	/// <summary>
	/// Locks the game loop mutex.
	/// </summary>
	static void lock();
	/// <summary>
	/// Unlocks the game loop mutex.
	/// </summary>
	static void unlock();
	/// <summary>
	/// Pause the game loop.
	/// </summary>
	static void pause();
	/// <summary>
	/// Unpause the game loop.
	/// </summary>
	static void unpause();

	/// Register the given eventHandler for time-based events. It will from now be called every <c>stepsize</c> seconds.
	/// </summary>
	/// <param name="eventHandler">The event handler.</param>
	/// <param name="stepsize">The stepsize in seconds.</param>
	static void registerForTimedEvents(std::function<void()>& eventHandler, double stepsize);
	/// Unregister the given eventHandler from time-based events.
	/// </summary>
	/// <param name="eventHandler">The event handler that was previously registered.</param>
	/// <param name="stepsize">The stepsize in seconds with which the handler was previously registered.</param>
	static void unregisterFromTimedEvents(std::function<void()>& eventHandler, double stepsize);

	/// Register the given eventHandler for frame-based events. It will from now be called every frame.
	/// </summary>
	/// <param name="eventHandler">The event handler.</param>
	static void registerForFrameEvents(std::function<void()>& eventHandler);
	/// Unregister the given eventHandler from frame-based events.
	/// </summary>
	/// <param name="eventHandler">The event handler that was previously registered.</param>
	static void unregisterFromFrameEvents(std::function<void()>& eventHandler);

	/// <summary>
	/// Gets the frame stamp of the currently running TeetairApplication.
	/// </summary>
	/// <returns>The osg FrameStamp.</returns>
	static osg::FrameStamp* getFrameStamp();

private:
	static GameLoop* instance();
	GameLoop();

	void loop();
	QTimer timer;
	TeetairApplication* app;
	QMutex frameMutex;

	QMultiMap<double, std::function<void()>*> timeListeners;
	std::unordered_map<double, int> timeListenerInvocations;
	void fireTimeListenerEvents();
	void fireTimeListenerEvents(double stepsize);

	std::unordered_set<std::function<void()>*> frameListeners;
	void fireFrameListenerEvents();

	std::queue<std::function<void()>> deferredListenerChanges;
	QMutex deferredListenerChangesMutex;

	bool paused;
};