#include "Jenga.h"

// For updating the new reached height
void Jenga::checkHeight(float height) {
	if (height > bestHeight) {
		bestHeight = height;
	int rounded = (int)((height + 10.f) * 100);
		heightHUD->setText("Height: " + std::to_string(rounded));
	}
}

// For the top view
osgViewer::View* create2DView(int x, int y, int width, int height,
	const osg::Vec3& lookDir, const osg::Vec3& up,
	osg::GraphicsContext* gc, osg::Node* scene)
{
	osg::ref_ptr<osgViewer::View> view = new osgViewer::View;
	view->getCamera()->setGraphicsContext(gc);
	view->getCamera()->setViewport(x, y, width, height);
	view->setSceneData(scene);

	osg::Vec3 center = scene->getBound().center();
	double radius = scene->getBound().radius();
	view->getCamera()->setViewMatrixAsLookAt(center - lookDir*(radius), center, up);
	view->getCamera()->setProjectionMatrixAsOrtho(-5, 5, -5, 5, 1.0f, 10000.0f);
	return view.release();
}

// Register the gestures for the menu
void Jenga::onConnect(const Leap::Controller& controller) {
	controller.enableGesture(Leap::Gesture::TYPE_CIRCLE);
	controller.enableGesture(Leap::Gesture::TYPE_SCREEN_TAP);
}

Jenga::Jenga(OsgViewerWidget* osgViewerWidget){
	GameLoop::lock();

	int width = 800, height = 600;
	osg::Vec4 backgroundColor = osg::Vec4(0.2, 0.2, 0.6, 1.0); //blue

	// Prepare camera
	osg::Camera* camera = getCamera();
	camera->setGraphicsContext(osgViewerWidget->getGraphicsContext());
	camera->setClearColor(backgroundColor);
	camera->setViewport(new osg::Viewport(0, 0, width, height));
	camera->setProjectionMatrixAsPerspective(30.0f, static_cast<double>(width) / static_cast<double>(height), 1.0f, 10000.0f);

	createScene();

	buildScene();
	
	// create the top view
	int leftPadding = 20, bottomPadding = 350;
	int topSize = 100;

	 top = create2DView(
		leftPadding, bottomPadding, topSize, topSize, -osg::Z_AXIS, osg::Y_AXIS, osgViewerWidget->getGraphicsContext(), towerScene.get());

	// Add the top view to the main view
	osgViewerWidget->addView(top.get());


	// Let the show begin!
	setSceneData(osgScene.get());
	
	// Add the controllable camera
	handCamera = new HandCamera;
	handCamera->setAngle(osg::PI);
	handCamera->setHeight(6.0);
	handCamera->setDistance(30);
	handCamera->setCenter(osg::Vec3(0, 0, -5));
	handCamera->setVerticalAxisFixed(true);
	setCameraManipulator(handCamera.get());

	// Add Leap support
	LeapManager::setListener(*this);
	osgViewerWidget->registerKeyListener(this);

	// Add all of the wood sounds
	for (int i = 1; i <= 10; i++)
		FM::addSound("hit" + std::to_string(i), "data/Jenga/wood-hit-" + std::to_string(i) + ".mp3");

	// Add menu hud
	MenuItemFolder* mainMenu = new MenuItemFolder("Menu");
	mainMenu->addItem(new MenuItemAction([&]() { resetScene(); }, "Restart"));
	mainMenu->addItem(new MenuItemAction([&]() { menu->hide(); }, "Continue"));
	
	menu = new MenuHUD();
	menu->attachToScene(osgScene, this);
	menu->setMainMenu(mainMenu);
	menu->setCurrentMenu(mainMenu);

	GameLoop::unlock();
}

void Jenga::createScene() {
	GameLoop::lock();
	// Prepare world
	osgScene = new osg::Group;

	towerScene = new osg::Group;

	osgScene->addChild(towerScene);
	
	sceneShader = new ShaderProgram(towerScene);
	sceneShader->setVertexShader("data/Jenga/standardShader.vert");
	sceneShader->setFragmentShader("data/Jenga/standardShader.frag");

	PX::createScene(osg::Vec3(0.0f, 0.0f, -9.81f));

	// Initialize Height HUD
	bestHeight = -10.f;

	heightHUD = new TextHUD();
	heightHUD->attachToScene(osgScene, this);
	heightHUD->setAlignment(osgText::TextBase::AlignmentType::RIGHT_TOP);
	heightHUD->setRelativePosition(osg::Vec2(0.9, 0.9));
	heightHUD->setColor(osg::Vec4(1, 1, 1, 1));
	heightHUD->setText("Height: 0");
	heightHUD->show();

	// Create first info HUD
	osg::ref_ptr<AnimatedTextHUD> infoTextHUD = new AnimatedTextHUD();
	infoTextHUD->attachToScene(osgScene, this);
	infoTextHUD->setColor(osg::Vec4(0, 1, 0, 1));
	infoTextHUD->setText("Try to stack the bricks as high as possible!");
	infoTextHUD->setTimes(10, 10, 5);
	infoTextHUD->show();

	GameLoop::unlock();
}

void Jenga::resetScene() {
	GameLoop::lock();
	clearScene();
	buildScene();
	GameLoop::unlock();
}

void Jenga::clearScene() {
	GameLoop::lock();
	// Remove everything in the scene
	towerScene->removeChildren(0, towerScene->getNumChildren());
	transforms.clear();
	PX::clearPhysxScene();
	GameLoop::unlock();
}

void Jenga::buildScene() {	
	GameLoop::lock();
	// Add ground
	osg::ref_ptr<osg::Geode> ground = new osg::Geode;
	osg::ref_ptr<osg::ShapeDrawable> groundShape = new osg::ShapeDrawable(new osg::Box(osg::Vec3(0.0, 0.0, -10.0), 30.0, 30.0, 0.1));
	groundShape->setColor(osg::Vec4(0.2, 0.2, 0.2, 1.0)); // Grey
	ground->addDrawable(groundShape.get());
	towerScene->addChild(ground.get());

	PX::createMaterial("standard", 0.5f, 0.5f, 0.f);
	PX::createPlane(GROUND, "standard", osg::Vec3(0.0, 0.0, -10.0), osg::Vec3(0.0, 0.0, 1.0), groups::PLANE, groups::BRICK);

	// Add tower
	osg::ref_ptr<osg::Geode> tower = new osg::Geode;
	osg::ref_ptr<osg::MatrixTransform> towerTransform = new osg::MatrixTransform(osg::Matrix::identity());

	towerTransform->getOrCreateStateSet()->addUniform(new Uniform("a_LightPos", osg::Vec3(1, 1, 1)));

	towerShader = new ShaderProgram(towerTransform.get());
	towerShader->setVertexShader("data/Jenga/brickShader.vert");
	towerShader->getVertexShader()->addCodeFileInjection("data/phong_vert.glsl", "//HOTSPOT_PRE_MAIN");
	towerShader->getVertexShader()->addCodeTextInjection("phong();", "//HOTSPOT_START_MAIN");

	towerShader->setFragmentShader("data/Jenga/brickShader.frag");
	towerShader->getFragmentShader()->addCodeFileInjection("data/phong_frag.glsl", "//HOTSPOT_PRE_MAIN");
	towerShader->getFragmentShader()->addCodeFileInjection("data/classicnoise3D.glsl", "//HOTSPOT_PRE_MAIN");
	towerShader->getFragmentShader()->addCodeFileInjection("data/wood.glsl", "//HOTSPOT_PRE_MAIN");

	int id = BRICK_START;
	float bias = 1.1; //for padding between the bricks
	// Add boxes
	for (int height = 0; height < 10; height++) {
		for (int i = 0; i < 3; i++) {

			osg::Vec3f position;
			osg::Vec3f dimension;
			if (height % 2 == 0) {
				position = osg::Vec3f(0, i*1.1 - 0.1, -10 + height * bias);
				dimension = osg::Vec3f(3.3, 1.0, 1.0);
			}
			else {
				position = osg::Vec3f((i - 1)*1.1, 1, -10 + height * bias);
				dimension = osg::Vec3f(1.0, 3.3, 1.0);
			}

			osg::ref_ptr<osg::Geode> brick = new osg::Geode;
			osg::ref_ptr<osg::ShapeDrawable> brickShape = new osg::ShapeDrawable(new osg::Box(osg::Vec3(), dimension.x(), dimension.y(), dimension.z()));

			osg::Vec4f color;
			int random = rand() % 4;
			switch (random) {
			case 0: color = Vec4f(1, 0, 0, 1); break;
			case 1: color = Vec4f(0, 1, 0, 1); break;
			case 2: color = Vec4f(0, 0, 1, 1); break;
			case 3: color = Vec4f(1, 1, 0, 1); break;
			default: color = Vec4f(1, 0, 0, 1); break;
			}

			brickShape->setColor(color);
			brick->addDrawable(brickShape);
			osg::ref_ptr<osg::MatrixTransform> brickTransform = new osg::MatrixTransform(osg::Matrix::translate(position));
			brickTransform->addChild(brick.get());

			brickTransform->getOrCreateStateSet()->addUniform(new Uniform("a_BrickId", id));

			towerTransform->addChild(brickTransform.get());
			PX::createBox(id, dimension, 0.1f, "standard", PX::ALL, groups::BRICK, groups::BRICK | groups::PLANE | groups::HAND);
			PX::setMatrix(id, brickTransform->getMatrix());
			// Register for sleep events
			PX::setSendSleepNotifies(id, true);
			PX::setSleepThreshold(id, 0.2);
			
			transforms[id] = brickTransform.get();

			id++;

		}
	}
	towerTransform->addChild(tower.get());
	towerScene->addChild(towerTransform.get());

	hand = new Gamehand((PhysxScene*)this, towerScene, groups::HAND, groups::BRICK);
	for (auto i : hand->transforms)
		transforms.insert(i);
	GameLoop::unlock();
}

void Jenga::onFrame(const Controller& controller) {
	GameLoop::lock();

	// Check for menu
	if (menu->isVisible()) {
		menu->onFrame(controller);
		GameLoop::unlock();
		return;
	}

	const Leap::Frame frame = controller.frame();
	
	// Check for hand camera control
	if (handCamera->handleLeapFrame(frame)) {
		double radius = osgScene.get()->getBound().radius();
		top->getCamera()->setViewMatrixAsLookAt(osg::Z_AXIS*(radius), osg::Vec3(), osg::Vec3(sin(osg::PI + handCamera->getAngle()), cos(osg::PI + handCamera->getAngle()), 0));
		GameLoop::unlock();
		return;
	}
	
	// Reset finger positions after 90 Frames (1,5 - 3 seconds) if 1 hand and 5 fingers are recognized.
	leapFingerResetCount++;
	if (leapFingerResetCount > 90 && frame.hands().count() == 1 && frame.fingers().count() == 5) {
		mustBeInitialized = true;
		leapFingerResetCount = 0;
	}
	
	if (frame.hands().count() == 0)
		mustBeInitialized = true;
	else if (frame.hands().count() == 1) {
		hand->synchronizeHand(frame, handCamera->getAngle(), mustBeInitialized);
		mustBeInitialized = false;
	}
	else {
		mustBeInitialized = true;
	}
	GameLoop::unlock();
}

void Jenga::keyReleaseEvent(QKeyEvent* event) {
	// Check for menu
	if (menu->isVisible()) {
		menu->keyReleaseEvent(event);
		return;
	}
	switch (event->key()) {
	case Qt::Key_R:
		resetScene();
		break;
	case Qt::Key_Escape:
		menu->toggleVisibility();
		break;
	default:
		return;
	}
	event->accept();
}

void Jenga::onSleep(PxActor **actors, PxU32 count) {
	// Check new height when a brick sleeps
	for (PxU32 i = 0; i < count; i++) {
		PxActor* actor = actors[i];
		for (int id = BRICK_START; id < BRICK_START + 30; id++) {
			if (PX::isActor(id, actor)) {
				osg::Matrix matrix = PX::getMatrix(id);
				checkHeight(matrix.getTrans().z());
			}
		}
	}
}

void Jenga::onContact(const physx::PxContactPairHeader& pairHeader, const physx::PxContactPair* pairs, physx::PxU32 nbPairs) {
	// Play sound when a brick is hit
	float strength = 0.0;
	for (PxU32 i = 0; i < nbPairs; i++)
	{
		PxContactPairPoint* contacts = new PxContactPairPoint[pairs[i].contactCount];
		PxU32 nbContacts = pairs[i].extractContacts(contacts, pairs[i].requiredBufferSize);
		for (PxU32 j = 0; j < nbContacts; j++)
		{
			PxVec3 point = contacts[j].position;
			PxVec3 impulse = contacts[j].impulse;
			strength = impulse.magnitude();
		}
		delete[] contacts;
	}
	if (strength > 0.01) {
		// Choose a random number between 1 and 10
		const unsigned long n = 10;
		const unsigned long divisor = (RAND_MAX + 1) / n;

		unsigned long k;
		do { k = std::rand() / divisor; } while (k >= n);

		// Play that random sound
		FM::playSound("hit" + std::to_string(k + 1), strength);
	}
}