#include "BlockButton.h"
#include "Teeter.h"
#include "Level.h"
#include "Board.h"
#include "Ball.h"

BlockButton::BlockButton(Level& level, float x, float y, int s, int b, std::vector<ref_ptr<Block>>& group)
: Block(level, x, y, s, b, group, true, "Button")
, activated(false)
, coolsDown(false)
{
	fixBlockOnBoard();

	if (s >= 91) { // "== 100"
		makeTrigger(s);
	}
	else if (s >= 59) { // ">= 60"
		makePushButton(s);
	}
	else if (s >= 49) { // "== 50"
		makeSwitch(s);
	}
	else { // fallback
		makeTrigger(s);
	}
	getLevel().getBoard()->registerContactableBlock(this, getId());

	buildButton();
}

void BlockButton::buildButton() {
	buttonGeode = new Geode();
	buttonGeode->addDrawable(new osg::ShapeDrawable(new Cylinder(Vec3(), 0.5, 0.3)));

	ref_ptr<StateSet> stateSet = buttonGeode->getOrCreateStateSet();
	buttonPose = new Uniform("a_Pose", Vec3f());
	stateSet->addUniform(buttonPose.get());
	buttonPoseAnimationTarget = new Uniform("a_PoseAnimationTarget", Vec3f());
	stateSet->addUniform(buttonPoseAnimationTarget.get());
	buttonPoseAnimationStart = new Uniform("a_PoseAnimationStart", 0.f);
	stateSet->addUniform(buttonPoseAnimationStart.get());
	buttonPoseAnimationEnd = new Uniform("a_PoseAnimationEnd", 0.f);
	stateSet->addUniform(buttonPoseAnimationEnd.get());
	buttonColor = new Uniform("a_Color", color);
	stateSet->addUniform(buttonColor.get());
	buttonColorAnimationTarget = new Uniform("a_ColorAnimationTarget", color);
	stateSet->addUniform(buttonColorAnimationTarget.get());
	buttonColorAnimationStart = new Uniform("a_ColorAnimationStart", 0.f);
	stateSet->addUniform(buttonColorAnimationStart.get());
	buttonColorAnimationEnd = new Uniform("a_ColorAnimationEnd", 0.f);
	stateSet->addUniform(buttonColorAnimationEnd.get());

	if (!getTeeter().shaders.count("animationShader.vert")) {
		getTeeter().shaders["animationShader.vert"] = new ShaderProgram::Shader("data/Teeter/animationShader.vert", Shader::VERTEX);
		getTeeter().shaders["animationShader.vert"]->addCodeFileInjection("data/easeFunctions.glsl", "//HOTSPOT_PRE_MAIN");
		getTeeter().shaders["animationShader.vert"]->addCodeFileInjection("data/phong_vert.glsl", "//HOTSPOT_PRE_MAIN");
	}
	if (!getTeeter().shaders.count("animationShader.frag")) {
		getTeeter().shaders["animationShader.frag"] = new ShaderProgram::Shader("data/Teeter/animationShader.frag", Shader::FRAGMENT);
		getTeeter().shaders["animationShader.frag"]->addCodeFileInjection("data/easeFunctions.glsl", "//HOTSPOT_PRE_MAIN");
		getTeeter().shaders["animationShader.frag"]->addCodeFileInjection("data/phong_frag.glsl", "//HOTSPOT_PRE_MAIN");	
	}

	buttonShader = new ShaderProgram(buttonGeode);
	buttonShader->setVertexShader(getTeeter().shaders["animationShader.vert"]);
	buttonShader->setFragmentShader(getTeeter().shaders["animationShader.frag"]);

	ref_ptr<MatrixTransform> buttonTransform = new MatrixTransform();
	Vec3 gradient = Vec3(1, 0, blockEndHeight - blockStartHeight);
	gradient.normalize();
	osg::Matrix tilt = osg::Matrix::rotate(gradient, Vec3(1, 0, 0));
	double rot = 0;
	switch (blockDirection)
	{
	case Block::GradientDirection::TOP:
		rot = osg::PI_2;
		break;
	case Block::GradientDirection::LEFT:
		rot = osg::PI;
		break;
	case Block::GradientDirection::BOTTOM:
		rot = -osg::PI_2;
		break;
	default:
		break;
	}
	osg::Matrix rotate = osg::Matrix::rotate(rot, Vec3(0, 0, 1));
	osg::Matrix translate = osg::Matrix::translate(0, 0, getHeight() / 2 - (blockEndHeight - blockStartHeight) / 2);
	buttonTransform->setMatrix(tilt * rotate * translate);
	buttonTransform->addChild(buttonGeode);
	getTransform()->addChild(buttonTransform);
}

void BlockButton::makeTrigger(int saturation) {
	type = Type::TRIGGER;
	color = Vec4f(1, 0, 0, 1);
}

void BlockButton::makePushButton(int saturation) {
	type = Type::PUSHBUTTON;
	color = Vec4f(0, 0, 1, 1);

	int offset = saturation - 60;
	coolDownDuration = (int)((offset + 3.0) / 3.0 + 0.5);
}

void BlockButton::makeSwitch(int saturation) {
	type = Type::SWITCH;
	color = Vec4f(1, 1, 0, 1);

	coolDownDuration = 1;
}

void BlockButton::contactedBy(Ball& ball) {
	if (coolsDown) {
		return;
	}

	coolsDown = true;
	triggerGroup();


	float grayscale = qGray(color.x() * 255, color.y() * 255, color.z() * 255) / 255.0f;
	animateColor(color, Vec4f(grayscale, grayscale, grayscale, grayscale), 0.5f);

	if (type == Type::TRIGGER) {
		setActivated(true);
	}
	else {
		setActivated(!activated);
		QTimer::singleShot(coolDownDuration * 1000, this, SLOT(cooledDown()));
	}
}

void BlockButton::setActivated(bool a) {
	activated = a;

	if (activated) {
		animatePose(Vec3f(), Vec3f(0, 0, -0.14), 0.5f);
	}
	else {
		animatePose(Vec3f(0, 0, -0.14), Vec3f(), 0.5f);
	}
}

void BlockButton::triggerGroup() {
	for (std::vector<ref_ptr<Block>>::iterator i = getGroup().begin(); i != getGroup().end(); ++i) {
		(*i)->triggeredBy(*this);
	}
}

void BlockButton::cooledDown() {
	float grayscale = qGray(color.x() * 255, color.y() * 255, color.z() * 255) / 255.0f;
	animateColor(Vec4f(grayscale, grayscale, grayscale, grayscale), color, 0.5f);

	if (type == Type::PUSHBUTTON) {
		setActivated(!activated);
		triggerGroup();
	}
	coolsDown = false;
}

void BlockButton::animateColor(Vec4f from, Vec4f to, float duration) {
	float time = GameLoop::getFrameStamp()->getSimulationTime();
	buttonColor->set(from);
	buttonColorAnimationTarget->set(to);
	buttonColorAnimationStart->set(time);
	buttonColorAnimationEnd->set(time + duration);
}

void BlockButton::animatePose(Vec3f from, Vec3f to, float duration) {
	float time = GameLoop::getFrameStamp()->getSimulationTime();
	buttonPose->set(from);
	buttonPoseAnimationTarget->set(to);
	buttonPoseAnimationStart->set(time);
	buttonPoseAnimationEnd->set(time + duration);
}

void BlockButton::reset() {
	activated = false;
	coolsDown = false;
	buttonPose->set(Vec3f());
	buttonPoseAnimationTarget->set(Vec3f());
	buttonColor->set(color);
	buttonColorAnimationTarget->set(color);
	std::cout << "reset" << std::endl;
}