#include "BlockCheckpoint.h"
#include "Teeter.h"
#include "Level.h"
#include "Board.h"
#include "Ball.h"

BlockCheckpoint::BlockCheckpoint(Level& level, float x, float y, int s, int b, std::vector<ref_ptr<Block>>& group)
: Block(level, x, y, s, b, group, true, "Checkpoint")
, mandatory(false)
, points(10)
, visited(false)
{
	fixBlockOnBoard();
	checkpoints.insert(this);
	
	if (s >= 91) { // "== 100"
		makeStart();
	}
	else if (s >= 79) { // ">= 80"
		makeCheckpoint(true, s - 80);
	}
	else if (s >= 69) { // ">= 70"
		makeCheckpoint(false, s - 70);
	}
	else if (s >= 49) { // ">= 50"
		makeEnd(s - 50);
	}
	else {
		makeCheckpoint(false, 0);
	}
	getLevel().getBoard()->registerContactableBlock(this, getId());
}

void BlockCheckpoint::makeStart() {
	type = Type::START;
	getLevel().setStartPoint(this);
}

void BlockCheckpoint::makeCheckpoint(bool m, int pointOffset) {
	type = Type::CHECKPOINT;
	mandatory = m;
	points = (int)((pointOffset + 3.0) / 3.0 + 0.5) * 10;
	createFlag();
	if (mandatory)
		flagGeode->getOrCreateStateSet()->addUniform(new Uniform("a_FlagColor", Vec4(0.8, 0, 0, 1)));
	else
		flagGeode->getOrCreateStateSet()->addUniform(new Uniform("a_FlagColor", Vec4(0, 0, 0.8, 1)));
}

void BlockCheckpoint::makeEnd(int pointOffset) {
	type = Type::END;
	points = (int)((pointOffset + 3.0) / 3.0 + 0.5) * 50;
	createFlag();
}

void BlockCheckpoint::contactedBy(Ball& ball) {
	switch (type) {
	case Type::CHECKPOINT:
		if (visited) {
			return;
		}
		visited = true;
		animatePose(Vec3f(), Vec3f(0.0f, 0.0f, -1.0f), 1.0f);
		getLevel().setLastCheckpoint(this);
		awardPoints();
		break;
	case Type::END:
		checkEnd();
		break;
	default:
		break;
	}
}

void BlockCheckpoint::checkEnd() {
	bool won = true;
	for (std::unordered_set<BlockCheckpoint*>::iterator i = checkpoints.begin(); i != checkpoints.end(); ++i) {
		if ((*i)->type == Type::CHECKPOINT && !(*i)->visited && (*i)->mandatory) {
			won = false;
			break;
		}
	}
	if (won) {
		awardPoints();
		getLevel().won();
	}
}

void BlockCheckpoint::awardPoints() {
	getLevel().changePoints(points);
}

void BlockCheckpoint::createFlag() {
	float groundHeight = getHeight() / 2 - (blockEndHeight - blockStartHeight) / 2;
	// create flag
	// initialize arrays
	osg::ref_ptr<osg::Vec3Array> vertices = new osg::Vec3Array(10);
	osg::Vec3 position = osg::Vec3(0.0f, 0.0f, groundHeight + 1.55f);
	for (unsigned int i = 0; i<10 - 1; i += 2)
	{
		(*vertices)[i] = position;
		(*vertices)[i + 1] = position - osg::Vec3(0, 0, 0.4);
		position += osg::Vec3(0.1, 0, 0);
	}

	// set geometry attributes
	osg::ref_ptr<osg::Geometry> geom = new osg::Geometry;
	geom->setDataVariance(osg::Object::STATIC);

	// bind the arrays to the geometry
	geom->setVertexArray(vertices.get());
	geom->addPrimitiveSet(new osg::DrawArrays(GL_QUAD_STRIP, 0, 10));

	flagGeode = new Geode();
	flagGeode->addDrawable(geom);
	ref_ptr<StateSet> stateSet = flagGeode->getOrCreateStateSet();
	stateSet->addUniform(new Uniform("a_FlagType", (int)type));
	flagPose = new Uniform("a_Pose", Vec3f());
	stateSet->addUniform(flagPose.get());
	flagPoseAnimationTarget = new Uniform("a_PoseAnimationTarget", Vec3f());
	stateSet->addUniform(flagPoseAnimationTarget.get());
	flagPoseAnimationStart = new Uniform("a_PoseAnimationStart", 0.f);
	stateSet->addUniform(flagPoseAnimationStart.get());
	flagPoseAnimationEnd = new Uniform("a_PoseAnimationEnd", 0.f);
	stateSet->addUniform(flagPoseAnimationEnd.get());

	getTransform()->addChild(flagGeode);

	flagShader = new ShaderProgram(flagGeode);
	if (!getTeeter().shaders.count("flagShader.vert")) {
		getTeeter().shaders["flagShader.vert"] = new ShaderProgram::Shader("data/Teeter/flagShader.vert", Shader::VERTEX);
		getTeeter().shaders["flagShader.vert"]->addCodeFileInjection("data/easeFunctions.glsl", "//HOTSPOT_PRE_MAIN");
	}
	if (!getTeeter().shaders.count("flagShader.frag")) {
		getTeeter().shaders["flagShader.frag"] = new ShaderProgram::Shader("data/Teeter/flagShader.frag", Shader::FRAGMENT);
	}
	flagShader->setVertexShader(getTeeter().shaders["flagShader.vert"]);
	flagShader->setFragmentShader(getTeeter().shaders["flagShader.frag"]);

	// create flagstaff
	flagstaffGeode = new Geode();
	flagstaffGeode->addDrawable(new osg::ShapeDrawable(new Cylinder(Vec3(0,0,groundHeight + 0.8), 0.03, 1.6)));
	getTransform()->addChild(flagstaffGeode);
}

void BlockCheckpoint::animatePose(Vec3f from, Vec3f to, float duration) {
	float time = GameLoop::getFrameStamp()->getSimulationTime();
	if (type != Type::START) {
		flagPose->set(from);
		flagPoseAnimationTarget->set(to);
		flagPoseAnimationStart->set(time);
		flagPoseAnimationEnd->set(time + duration);
	}
}

void BlockCheckpoint::reset() {
	visited = false;
	animatePose(Vec3f(0.0f, 0.0f, -1.0f), Vec3f(), 1.0f);
}

BlockCheckpoint::~BlockCheckpoint() {
	checkpoints.erase(this);
}

std::unordered_set<BlockCheckpoint*> BlockCheckpoint::checkpoints = std::unordered_set<BlockCheckpoint*>();