#include "BlockElevator.h"
#include "Teeter.h"
#include "Level.h"
#include "Board.h"

BlockElevator::BlockElevator(Level& level, float x, float y, int s, int b, std::vector<ref_ptr<Block>>& group)
: Block(level, x, y, s, b, group, true, "Elevator")
, direction(1)
, autoMovement(false)
, fixed(false)
, timer(this)
{
	int offset = s;
	if (s >= 74) { // ">= 75"
		offset -= 75;
	}
	else { // ">= 50" and fallback
		autoMovement = true;
		offset -= 50;
	}
	offset = max(0, offset);
	offset = ((int)(offset / 3.0 + 0.5)) * 3;

	if (offset <= 9) {
		offset -= 12; // negative movements, 0 = -4, 3 = -3 etc.
	}
	else {
		offset -= 9; // positive movements, 12 = 1, 15 = 2 etc.
	}

	movement = offset / 3 / 2.0;

	// create joint based on movement
	joint = getTeeter().PX::getNextJointId();
	getTeeter().PX::createD6Joint(joint, getId(), osg::Matrix::identity(), getLevel().getBoard()->getId(), osg::Matrix::translate(0, 0, movement / 2.0) * osg::Matrix::translate(getTransform()->getMatrix().getTrans()));
	getTeeter().PX::setD6Drive(joint, PxD6Drive::eZ, 20, 0, 10, true);
	// create fixed joint
	fixedJoint = getTeeter().PX::getNextJointId();

	
	timer.setSingleShot(true);
	timer.setInterval(1000);
	connect(&timer, &QTimer::timeout, this, &BlockElevator::move);

	if (autoMovement) {
		timer.start();
	}

	frameFunction = std::bind(&BlockElevator::frameEvent, this);
	GameLoop::registerForFrameEvents(frameFunction);
	
	reset();
}

void BlockElevator::triggeredBy(Block& block) {
	move();
}

void BlockElevator::frameEvent() {
	if (fixed)
		return;
	osg::Vec3 posInBoard(getX(), getY(), getHeight() / 2.0);
	osg::Vec3 dynamicPos = Vec3() * getTransform()->getMatrix();
	osg::Vec3 staticPos = posInBoard * getLevel().getBoard()->getTransform()->getMatrix();
	double length = (int)((dynamicPos - staticPos).length() * 100 + 0.5) / 100.0;
	if (length + 0.01 >= abs(movement) && direction == 1) {
		getTeeter().PX::createFixedJoint(fixedJoint, getId(), osg::Matrix::identity(), getLevel().getBoard()->getId(), osg::Matrix::translate(getX(), getY(), getHeight() / 2.0 + movement));
		fixed = true;
		if (autoMovement) {
			timer.start();
		}
	}

	if (length - 0.01 <= 0 && direction == -1) {
		getTeeter().PX::createFixedJoint(fixedJoint, getId(), osg::Matrix::identity(), getLevel().getBoard()->getId(), osg::Matrix::translate(getX(), getY(), getHeight() / 2.0));
		fixed = true;
		if (autoMovement) {
			timer.start();
		}
	}
}

void BlockElevator::move() {
	timer.stop();
	direction *= -1;
	osg::Vec3 local = Vec3(0, 0, direction * (-movement));
	getTeeter().PX::setD6Motion(joint, PxD6Axis::eZ, PxD6Motion::eLIMITED, abs(movement / 2.0), 0);
	getTeeter().PX::d6Drive(joint, osg::Matrix::translate(local), Vec3(0, 0, 0.1), Vec3());
	if (fixed) {
		getTeeter().PX::releaseJoint(fixedJoint);
		fixed = false;
	}
}

void BlockElevator::reset() {
	direction = 1;
	move();
}

BlockElevator::~BlockElevator() {
	GameLoop::unregisterFromFrameEvents(frameFunction);
}