#include "BlockPortal.h"
#include "Teeter.h"
#include "Level.h"
#include "Board.h"
#include "Ball.h"

BlockPortal::BlockPortal(Level& level, float x, float y, int s, int b, std::vector<ref_ptr<Block>>& group)
: Block(level, x, y, s, b, group, true, "Portal")
{ 
	fixBlockOnBoard();

	if (s >= 89) { // ">= 90"
		orientation = Orientation::RIGHT;
		makePortal(offset(s, 90));
	}
	else if (s >= 79) { // ">= 80"
		orientation = Orientation::LEFT;
		makePortal(offset(s, 80));
	}
	else if(s >= 69) { // ">= 70"
		orientation = Orientation::BACK;
		makePortal(offset(s, 70));
	}
	else if(s >= 59) { // ">= 60"
		orientation = Orientation::FRONT;
		makePortal(offset(s, 60));
	}
	else if(s >= 49) { // ">= 50"
		orientation = Orientation::TOP;
		makePortal(offset(s, 50));
	}
	else { // fallback
		orientation = Orientation::RIGHT;
		makePortal(0);
	}
}

int BlockPortal::offset(int saturation, int baseSaturation) {
	int offset = max(0, saturation - baseSaturation);
	offset = ((int)(offset / 3.0 + 0.5)) * 3;
	return offset;
}

void BlockPortal::makePortal(int offset) {
	makeTrigger();
	frameFunction = std::bind(&BlockPortal::frameEvent, this);
	attraction = offset / 6.0;
	partner = nullptr;
	progress = Progress::WAIT;


	createParticles();

	ref_ptr<Geode> portalGeode = new Geode();
	// Render a cylinder as the portal entrance
	cylinder = new Cylinder(entranceCenter, 0.4, 0.5);
	// Rotate it according to the direction
	osg::Quat quat = osg::Quat();
	quat.makeRotate(osg::Vec3(0, 0, 1), -entranceDirection);
	cylinder->setRotation(quat);
	// Make it blue
	ref_ptr<ShapeDrawable> drawable = new osg::ShapeDrawable(cylinder);
	drawable->setColor(osg::Vec4(0, 0, 0.8, 1));
	// Set the standard shader
	ShaderProgram* portalShader = new ShaderProgram(portalGeode);
	portalShader->setVertexShader("data/Teeter/standardShader.vert");
	portalShader->setFragmentShader("data/Teeter/standardShader.frag");
	portalGeode->addDrawable(drawable);
	getTransform()->addChild(portalGeode);
}

void BlockPortal::createParticles() {
	
	// Create and initialize a particle system
	osgParticle::ParticleSystem *particleSystem = new osgParticle::ParticleSystem;

	// Set corresponding state set
	osg::StateSet *stateset = new osg::StateSet;
	stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
	osg::BlendFunc *blend = new osg::BlendFunc;
	blend->setFunction(osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE);
	stateset->setAttributeAndModes(blend, osg::StateAttribute::ON);
	particleSystem->setStateSet(stateset);

	
	osg::Geode *geode = new osg::Geode;

	shader = new ShaderProgram(geode);
	shader->setVertexShader("data/Teeter/standardShader.vert");
	shader->setFragmentShader("data/Teeter/standardShader.frag");

	getTransform()->addChild(geode);

	geode->addDrawable(particleSystem);

	// Add an 'updater' to help per-frame management
	osgParticle::ParticleSystemUpdater *particleSystemUpdater = new osgParticle::ParticleSystemUpdater;

	// Associate this updater with our particle system 
	particleSystemUpdater->addParticleSystem(particleSystem);
	// add the updater node to the scene graph
	getTransform()->addChild(particleSystemUpdater);

	// Create a particle to be used by our particle system and define a few of its properties
	osgParticle::Particle portalParticle;
	portalParticle.setSizeRange(osgParticle::rangef(0.01, 0.02));
	portalParticle.setLifeTime(1.0);
	portalParticle.setMass(0.01);

	
	portalParticle.setColorRange(osgParticle::rangev4(osg::Vec4f(1.0, 1.0, 0.0, 1.0), osg::Vec4f(1.0, 0.0, 0.0, 1.0)));

	// Make this our particle system's default particle 
	particleSystem->setDefaultParticleTemplate(portalParticle);

	// Create a modular emitter (this contains default counter, placer and shooter.)
	osgParticle::ModularEmitter *emitter = new osgParticle::ModularEmitter;

	// Associate this emitter with the particle system 
	emitter->setParticleSystem(particleSystem);

	// Get a handle to the emitter's counter and adjust the number of new particles
	// that will be added each frame
	osgParticle::RandomRateCounter *dustRate = static_cast<osgParticle::RandomRateCounter *>(emitter->getCounter());
	dustRate->setRateRange(50, 100); // generate 50 to 100 particles per second

	osgParticle::MultiSegmentPlacer* lineSegment = new osgParticle::MultiSegmentPlacer();

	lineSegment->addVertex(entranceCenter - entranceDirection);
	lineSegment->addVertex(entranceCenter - entranceDirection + osg::Vec3(0.01, 0.01, 0.01));
	// Use this placer for our emitter 
	emitter->setPlacer(lineSegment);

	// To customize the shooter, create and initialize a radial shooter 
	osgParticle::RadialShooter* shooter = new osgParticle::RadialShooter();

	// Set properties of this shooter 
	shooter->setThetaRange(0.0, osg::PI);
	shooter->setInitialSpeedRange(1, 2);
	// Use this shooter for our emitter
	emitter->setShooter(shooter);


	osgParticle::ModularProgram *program = new osgParticle::ModularProgram;
	program->setParticleSystem(particleSystem);
	getTransform()->addChild(program);

	osgParticle::FluidFrictionOperator *airFriction = new osgParticle::FluidFrictionOperator;
	airFriction->setFluidToAir();
	airFriction->setWind(entranceDirection);
	program->addOperator(airFriction);

	getTransform()->addChild(emitter);
}

void BlockPortal::makeTrigger() {
	float bl = (blockDirection == GradientDirection::TOP || blockDirection == GradientDirection::RIGHT ? blockEndHeight : blockStartHeight);
	float br = (blockDirection == GradientDirection::TOP || blockDirection == GradientDirection::LEFT ? blockEndHeight : blockStartHeight);
	float tl = (blockDirection == GradientDirection::BOTTOM || blockDirection == GradientDirection::RIGHT ? blockEndHeight : blockStartHeight);
	float tr = (blockDirection == GradientDirection::BOTTOM || blockDirection == GradientDirection::LEFT ? blockEndHeight : blockStartHeight);
	
	// The inside trigger
	ref_ptr<Vec3Array> innerVertices = new Vec3Array;
	innerVertices->push_back(Vec3(-0.4, -0.4, 0.1));		// bl bottom	(0)
	innerVertices->push_back(Vec3(-0.4, -0.4, bl - 0.1));	// bl top		(1)
	innerVertices->push_back(Vec3(-0.4,  0.4, 0.1));		// tl bottom	(2)
	innerVertices->push_back(Vec3(-0.4,  0.4, tl - 0.1));	// tl top		(3)
	innerVertices->push_back(Vec3( 0.4, -0.4, 0.1));		// br bottom	(4)
	innerVertices->push_back(Vec3( 0.4, -0.4, br - 0.1));	// br top		(5)
	innerVertices->push_back(Vec3( 0.4,  0.4, 0.1));		// tr bottom	(6)
	innerVertices->push_back(Vec3( 0.4,  0.4, tr - 0.1));	// tr top		(7)
	
	// BlockPortals are always standalone (see Block::buildConvexBlock)
	for (Vec3Array::iterator i = innerVertices->begin(); i != innerVertices->end(); ++i) {
		(*i) += Vec3(0, 0, -getHeight() / 2.0f);
	}

	getTeeter().PX::addConvexBoxToCompound(getId(), *innerVertices.get(), "holz", Level::Groups::BOARD, Level::Groups::BALL, true);

	// The outside trigger
	ref_ptr<Vec3Array> entranceVertices = new Vec3Array;
	switch (orientation)
	{
	case BlockPortal::Orientation::RIGHT:
		entranceVertices->push_back(Vec3(0.3, -0.4, br - 0.1));				// bl top
		entranceVertices->push_back(Vec3(0.3, -0.4, max(br, tr) - 0.8));	// bl bottom
		entranceVertices->push_back(Vec3(0.3,  0.4, tr - 0.1));				// tl top
		entranceVertices->push_back(Vec3(0.3,  0.4, max(br, tr) - 0.8));	// tl bottom
		entranceDirection = Vec3(-1, 0, 0);
		break;
	case BlockPortal::Orientation::LEFT:
		entranceVertices->push_back(Vec3(-0.3, -0.4, bl - 0.1));			// br top
		entranceVertices->push_back(Vec3(-0.3, -0.4, max(bl, tl) - 0.8));	// br bottom
		entranceVertices->push_back(Vec3(-0.3,  0.4, tl - 0.1));			// tr top
		entranceVertices->push_back(Vec3(-0.3,  0.4, max(bl, tl) - 0.8));	// tr bottom
		entranceDirection = Vec3(1, 0, 0);
		break;
	case BlockPortal::Orientation::BACK:
		entranceVertices->push_back(Vec3(-0.4, 0.3, tl - 0.1));				// bl top
		entranceVertices->push_back(Vec3(-0.4, 0.3, max(tl, tr) - 0.8));	// bl bottom
		entranceVertices->push_back(Vec3( 0.4, 0.3, tr - 0.1));				// br top
		entranceVertices->push_back(Vec3( 0.4, 0.3, max(tl, tr) - 0.8));	// br bottom
		entranceDirection = Vec3(0, -1, 0);
		break;
	case BlockPortal::Orientation::FRONT:
		entranceVertices->push_back(Vec3(-0.4, -0.3, bl - 0.1));			// tl top
		entranceVertices->push_back(Vec3(-0.4, -0.3, max(bl, br) - 0.8));	// tl bottom
		entranceVertices->push_back(Vec3( 0.4, -0.3, br - 0.1));			// tr top
		entranceVertices->push_back(Vec3( 0.4, -0.3, max(bl, br) - 0.8));	// tr bottom
		entranceDirection = Vec3(0, 1, 0);
		break;
	case BlockPortal::Orientation::TOP:
	default:
		entranceVertices->push_back(Vec3(-0.4,  0.4, tl - 0.2));			// tl top
		entranceVertices->push_back(Vec3( 0.4,  0.4, tr - 0.2));			// tr top
		entranceVertices->push_back(Vec3(-0.4, -0.4, bl - 0.2));			// bl top
		entranceVertices->push_back(Vec3( 0.4, -0.4, br - 0.2));			// br top
		entranceDirection = Vec3(0, 0, -1);
		break;
	}

	ref_ptr<Vec3Array> outerVertices = new Vec3Array;
	entranceCenter = Vec3();
	
	for (Vec3Array::iterator i = entranceVertices->begin(); i != entranceVertices->end(); ++i) {
		outerVertices->push_back(*i);
		outerVertices->push_back(*i - entranceDirection * 1.7);
		entranceCenter += *i;
	}
	entranceCenter /= 4;

	// BlockPortals are always standalone (see Block::buildConvexBlock)
	for (Vec3Array::iterator i = outerVertices->begin(); i != outerVertices->end(); ++i) {
		(*i) += Vec3(0, 0, -getHeight() / 2.0f);
	}
	entranceCenter += Vec3(0, 0, -getHeight() / 2.0f);

	getTeeter().PX::addConvexBoxToCompound(getId(), *outerVertices.get(), "holz", Level::Groups::BOARD, Level::Groups::BALL, true);
	getLevel().getBoard()->registerTriggerableBlock(this, getId());
}

void BlockPortal::triggeredBy(Ball& ball) {
	switch (progress)
	{
	case Progress::WAIT:
		startAttract();
		break;
	case Progress::ATTRACT:
		startSend();
		break;
	default:
		break;
	}
}

void BlockPortal::untriggeredBy(Ball& ball) {
	switch (progress)
	{
	case Progress::ATTRACT:
	case Progress::REPEL:
		reset();
		break;
	default:
		break;
	}
}

void BlockPortal::startAttract() {
	progress = Progress::ATTRACT;
	
	getTeeter().PX::setSimulateCollisions(getId(), false);
	GameLoop::registerForFrameEvents(frameFunction);
}

void BlockPortal::attractBall() {
	Vec3 ballP = Vec3() * getLevel().getBall()->getTransform()->getMatrix();
	Vec3 attractionCenterP = (entranceCenter + entranceDirection*0.5) * getTransform()->getMatrix();
	Vec3 attractionDirection = (attractionCenterP - ballP);

	getTeeter().PX::addForce(getLevel().getBall()->getId(), attractionDirection * attraction);
}

void BlockPortal::startSend() {
	partner = randomPartner();
	if (partner == nullptr) {
		return;
	}

	progress = Progress::SEND;
	getTeeter().PX::setForceResponsiveness(getLevel().getBall()->getId(), PhysxScene::ForceResponsiveness::NO_GRAVITY);
	getTeeter().PX::addForce(getLevel().getBall()->getId(), getTransform()->getMatrix().getRotate() * (entranceDirection)* 3);

	ref_ptr<StateSet> stateSet = getLevel().getBall()->getTransform()->getOrCreateStateSet();
	stateSet->getUniform("a_PortalActive")->set(true);
	if (!stateSet->getUniform("a_PortalSenderMatrix")) {
		stateSet->addUniform(new Uniform("a_PortalSenderMatrix", Matrixf()));
		stateSet->addUniform(new Uniform("a_PortalReceiverMatrix", Matrixf()));
		stateSet->addUniform(new Uniform("a_PortalRotationMatrix", Matrixf()));
	}
	stateSet->getUniform("a_PortalSenderMatrix")->setUpdateCallback(new BallPortalShaderCallback(this));
	stateSet->getUniform("a_PortalReceiverMatrix")->setUpdateCallback(new BallPortalShaderCallback(partner));
	stateSet->getUniform("a_PortalRotationMatrix")->set(Matrixf::rotate(entranceDirection, -partner->entranceDirection));

	partner->startReceive();
}

BlockPortal* BlockPortal::randomPartner() {
	std::vector<ref_ptr<BlockPortal>> partners;
	for (std::vector<ref_ptr<Block>>::iterator i = getGroup().begin(); i != getGroup().end(); ++i) {
		BlockPortal* block = dynamic_cast<BlockPortal*>((*i).get());
		if (block != nullptr && block != this) {
			partners.push_back(block);
		}
	}
	if (partners.empty())
		return nullptr;

	const unsigned long n = std::distance(partners.begin(), partners.end());
	const unsigned long divisor = (RAND_MAX + 1) / n;

	unsigned long k;
	do { k = std::rand() / divisor; } while (k >= n);

	return partners[k];
}

void BlockPortal::sendBall() {
	Vec3 entranceCenterP = entranceCenter * getTransform()->getMatrix();
	Vec3 ballP = Vec3() * getLevel().getBall()->getTransform()->getMatrix();
	if ((ballP - entranceCenterP).length() > 0.8) {
		reset();
		return;
	}
	float relPos = (ballP - entranceCenterP) * entranceDirection;
	if (relPos > 0.1) {
		sendComplete();
		return;
	}
}

void BlockPortal::startReceive() {
	progress = Progress::RECEIVE;

	getTeeter().PX::setSimulateCollisions(getId(), false);
}

void BlockPortal::sendComplete() {
	osg::Matrix senderMatrix = getTransform()->getMatrix();
	osg::Matrix receiverMatrix = osg::Matrix::translate(-partner->entranceDirection)*partner->getTransform()->getMatrix();
	osg::Matrix rotationMatrix = osg::Matrix::rotate(entranceDirection, -partner->entranceDirection);

	osg::Matrix newPose = getLevel().getBall()->getTransform()->getMatrix()*osg::Matrix::inverse(senderMatrix) * rotationMatrix * receiverMatrix;	

	getTeeter().PX::setMatrix(getLevel().getBall()->getId(), newPose);

	Vec3 oldVel = getTeeter().PX::getLinearVelocity(getLevel().getBall()->getId());
	getTeeter().PX::setLinearVelocity(getLevel().getBall()->getId(), oldVel * rotationMatrix);

	partner->startRepel();
	reset();
}

void BlockPortal::startRepel() {
	progress = Progress::REPEL;

	getTeeter().PX::addForce(getLevel().getBall()->getId(), getTransform()->getMatrix().getRotate() * (-entranceDirection) * 2);
}


void BlockPortal::frameEvent() {
	switch (progress)
	{
	case BlockPortal::Progress::ATTRACT:
		attractBall();
		break;
	case BlockPortal::Progress::SEND:
		sendBall();
		break;
	default:
		break;
	}
}

void BlockPortal::reset() {
	getTeeter().PX::setSimulateCollisions(getId(), true);
	getTeeter().PX::setForceResponsiveness(getLevel().getBall()->getId(), PhysxScene::ForceResponsiveness::ALL);
	getLevel().getBall()->getTransform()->getOrCreateStateSet()->getUniform("a_PortalActive")->set(false);
	partner = nullptr;
	if (progress != Progress::WAIT) {
		GameLoop::unregisterFromFrameEvents(frameFunction);
	}
	progress = Progress::WAIT;
}

BlockPortal::~BlockPortal() {
	if (progress != Progress::WAIT) {
		GameLoop::unregisterFromFrameEvents(frameFunction);
	}
}

void BlockPortal::BallPortalShaderCallback::operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) {
	if (portal->progress == Progress::RECEIVE) {
		uniform->set(Matrixf(Matrixf::translate(-portal->entranceDirection) * portal->getTransform()->getMatrix()));
	}
	else {
		uniform->set(Matrixf(portal->getTransform()->getMatrix()));
	}
}