#include "BlockTrap.h"
#include "Teeter.h"
#include "Level.h"
#include "Board.h"
#include "Ball.h"

BlockTrap::BlockTrap(Level& level, float x, float y, int s, int b, std::vector<ref_ptr<Block>>& group)
: Block(level, x, y, s, b, group, true, "Trap")
, timer()
{ 
	fixBlockOnBoard();

	if (s >= 89) { // ">= 90"
		makeMine(offset(s, 90));
	}
	else if (s >= 79) { // ">= 80"
		makeSpring(offset(s, 80));
	}
	else if(s >= 69) { // ">= 70"
		makeTrapdoor(offset(s, 70));
	}
	else if(s >= 59) { // ">= 60"
		// empty
	}
	else if(s >= 49) { // ">= 50"
		makeBlackHole(offset(s, 50));
	}
	else { // fallback
		makeMine(0);
	}
}

int BlockTrap::offset(int saturation, int baseSaturation) {
	int offset = max(0, saturation - baseSaturation);
	offset = ((int)(offset / 3.0 + 0.5)) * 3;
	return offset;
}


void BlockTrap::makeMine(int offset) {
	type = Type::MINE;
	delay = offset / 3;
	triggered = false;
	exploded = false;
	getLevel().getBoard()->registerContactableBlock(this, getId());


	ref_ptr<Geode> mineGeode = new Geode();
	float groundHeight = getHeight() / 2 - (blockEndHeight - blockStartHeight) / 2;
	sphere = new Sphere(osg::Vec3(0, 0, groundHeight + 0.1), 0.1);
	ref_ptr<ShapeDrawable> drawable = new osg::ShapeDrawable(sphere);
	drawable->setColor(osg::Vec4(1, 0, 0, 1));
	ShaderProgram* mineShader = new ShaderProgram(mineGeode);
	mineShader->setVertexShader("data/Teeter/mineShader.vert");
	mineShader->setFragmentShader("data/Teeter/mineShader.frag");
	mineShader->getVertexShader()->addCodeFileInjection("data/phong_vert.glsl", "//HOTSPOT_PRE_MAIN");
	mineShader->getVertexShader()->addCodeTextInjection("phong();", "//HOTSPOT_START_MAIN");
	mineShader->getFragmentShader()->addCodeFileInjection("data/phong_frag.glsl", "//HOTSPOT_PRE_MAIN");
	mineGeode->addDrawable(drawable);
	getTransform()->addChild(mineGeode);


}

void BlockTrap::triggerMine() {
	if (triggered)
		return;
	triggered = true;
	if (delay > 0) {
		singleShot(delay * 1000, &BlockTrap::mineExplode);
	}
	else {
		mineExplode();
	}
}

void BlockTrap::hideMine() {
	sphere->setRadius(0);
	singleShot(10000, &BlockTrap::reset);
}

void BlockTrap::mineExplode() {
	osg::Vec3 minePos = Vec3() * getTransform()->getMatrix();
	osg::Vec3 ballPos = Vec3() * getLevel().getBall()->getTransform()->getMatrix();
	double distance = (minePos - ballPos).length();
	getTeeter().PX::addForce(getLevel().getBall()->getId(), (ballPos - minePos) / distance * 100 / (distance * distance));
	getLevel().changeLives(-5.0 / (distance * distance), false);
	exploded = true;
	singleShot(500, &BlockTrap::hideMine);
	sphere->setRadius(0.5);
}

void BlockTrap::resetMine() {
	triggered = false;
	exploded = false;
	sphere->setRadius(0.1);
}

void BlockTrap::makeSpring(int offset) {
	type = Type::SPRING;
	strength = offset + 5;
	triggered = false;
	getLevel().getBoard()->registerContactableBlock(this, getId());

	// create spring
	float groundHeight = getHeight() / 2 - (blockEndHeight - blockStartHeight) / 2;

	ref_ptr<Geode> flagstaffGeode = new Geode();
	cylinder = new Cylinder(Vec3(0, 0, groundHeight), 0.4, 0.3);
	flagstaffGeode->addDrawable(new osg::ShapeDrawable(cylinder));
	getTransform()->addChild(flagstaffGeode);
}

void BlockTrap::triggerSpring() {
	if (triggered)
		return;
	triggered = true;
	cylinder->setHeight(1);
	osg::Vec3 springDirection = Vec3(0, 0, 1) * getLevel().getBoard()->getTransform()->getMatrix();
	getTeeter().PX::addForce(getLevel().getBall()->getId(), springDirection * strength * 2);
	singleShot(10000, &BlockTrap::reset);
}

void BlockTrap::resetSpring() {
	triggered = false;
	cylinder->setHeight(0.3);
}

void BlockTrap::makeTrapdoor(int offset) {
	type = Type::TRAPDOOR;
	delay = offset / 3;
	triggered = false;
	open = false;
	callback = new BlockTrapdoorCallback(this);
	getTransform()->setCullCallback(callback);
	getLevel().getBoard()->registerContactableBlock(this, getId());
}

void BlockTrap::triggerTrapdoor() {
	if (triggered)
		return;
	triggered = true;
	if (delay > 0) {
		singleShot(delay * 1000, &BlockTrap::trapdoorOpen);
	}
	else {
		trapdoorOpen();
	}
}

void BlockTrap::trapdoorOpen() {
	open = true;
	getTeeter().PX::setSimulateCollisions(getId(), false);
	singleShot(10000, &BlockTrap::reset);
}

void BlockTrap::resetTrapdoor() {
	triggered = false;
	open = false;
	getTeeter().PX::setSimulateCollisions(getId(), true);
}

void BlockTrap::makeBlackHole(int offset) {
	type = Type::BLACK_HOLE;
	strength = offset / 2 + 3;
	getTeeter().PX::addBoxToCompound(getId(), Vec3(strength, strength, strength), "holz", osg::Matrix::identity(), Level::Groups::BOARD, Level::Groups::BALL, true);
	frameFunction = std::bind(&BlockTrap::frameEvent, this);
	entered = false;
	getLevel().getBoard()->registerTriggerableBlock(this, getId());

	createBlackHoleParticles();

	ref_ptr<Geode> blackHoleGeode = new Geode();
	float groundHeight = getHeight() / 2 - (blockEndHeight - blockStartHeight) / 2;
	cylinder = new Cylinder(Vec3(0, 0, groundHeight), 0.2 + 0.1 * strength, 0.01);
	ref_ptr<ShapeDrawable> drawable = new osg::ShapeDrawable(cylinder);
	drawable->setColor(osg::Vec4(0, 0, 0, 1));
	ShaderProgram* blackHoleShader = new ShaderProgram(blackHoleGeode);
	blackHoleShader->setVertexShader("data/Teeter/standardShader.vert");
	blackHoleShader->setFragmentShader("data/Teeter/standardShader.frag");
	blackHoleGeode->addDrawable(drawable);
	getTransform()->addChild(blackHoleGeode);
}

void BlockTrap::createBlackHoleParticles() {

	// Create and initialize a particle system
	osgParticle::ParticleSystem *particleSystem = new osgParticle::ParticleSystem;

	// Set corresponding state set
	osg::StateSet *stateset = new osg::StateSet;
	stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
	osg::BlendFunc *blend = new osg::BlendFunc;
	blend->setFunction(osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE);
	stateset->setAttributeAndModes(blend, osg::StateAttribute::ON);
	particleSystem->setStateSet(stateset);


	osg::Geode *geode = new osg::Geode;

	shader = new ShaderProgram(geode);
	shader->setVertexShader("data/Teeter/standardShader.vert");
	shader->setFragmentShader("data/Teeter/standardShader.frag");

	getTransform()->addChild(geode);

	geode->addDrawable(particleSystem);

	// Add an 'updater' to help per-frame management
	osgParticle::ParticleSystemUpdater *particleSystemUpdater = new osgParticle::ParticleSystemUpdater;

	// Associate this updater with our particle system 
	particleSystemUpdater->addParticleSystem(particleSystem);
	// add the updater node to the scene graph
	getTransform()->addChild(particleSystemUpdater);

	// Create a particle to be used by our particle system and define a few of its properties
	osgParticle::Particle particle;
	particle.setSizeRange(osgParticle::rangef(0.03, 0.07));
	particle.setLifeTime(1.0);
	particle.setMass(0.01);
	particle.setShape(osgParticle::Particle::Shape::HEXAGON);

	particle.setColorRange(osgParticle::rangev4(osg::Vec4f(0.0, 0.0, 0.0, 1.0), osg::Vec4f(1.0, 0.0, 0.0, 1.0)));

	// Make this our particle system's default particle 
	particleSystem->setDefaultParticleTemplate(particle);

	// Create a modular emitter (this contains default counter, placer and shooter.)
	osgParticle::ModularEmitter *emitter = new osgParticle::ModularEmitter;

	// Associate this emitter with the particle system 
	emitter->setParticleSystem(particleSystem);

	// Get a handle to the emitter's counter and adjust the number of new particles
	// that will be added each frame
	osgParticle::RandomRateCounter *dustRate = static_cast<osgParticle::RandomRateCounter *>(emitter->getCounter());
	dustRate->setRateRange(100, 200); // generate 100 to 200 particles per second

	osgParticle::MultiSegmentPlacer* lineSegment = new osgParticle::MultiSegmentPlacer();

	lineSegment->addVertex(-0.5, -0.5, 1);
	lineSegment->addVertex(-0.5, 0.5, 1);
	lineSegment->addVertex(0.5, 0.5, 1);
	lineSegment->addVertex(0.5, -0.5, 1);
	// Use this placer for our emitter 
	emitter->setPlacer(lineSegment);

	// To customize the shooter, create and initialize a radial shooter 
	osgParticle::RadialShooter* shooter = new osgParticle::RadialShooter();

	// Set properties of this shooter 
	shooter->setThetaRange(0.0, osg::PI);
	shooter->setInitialSpeedRange(1, 2);
	// Use this shooter for our emitter
	emitter->setShooter(shooter);


	osgParticle::ModularProgram *program = new osgParticle::ModularProgram;
	program->setParticleSystem(particleSystem);
	getTransform()->addChild(program);

	osgParticle::FluidFrictionOperator *airFriction = new osgParticle::FluidFrictionOperator;
	airFriction->setFluidToAir();
	airFriction->setWind(osg::Vec3(0, 0, -1));
	program->addOperator(airFriction);

	getTransform()->addChild(emitter);
}

void BlockTrap::triggerBlackHole() {
	entered = true;
	GameLoop::registerForFrameEvents(frameFunction);
}

void BlockTrap::blackHoleAttract() {
	if (!entered)
		return;
	osg::Vec3 holePos = Vec3() * getTransform()->getMatrix();
	osg::Vec3 ballPos = Vec3() * getLevel().getBall()->getTransform()->getMatrix();
	double distance = (holePos - ballPos).length();
	getTeeter().PX::addForce(getLevel().getBall()->getId(), (holePos - ballPos) / distance / (distance * distance));
}

void BlockTrap::untriggerBlackHole() {
	entered = false;
	GameLoop::unregisterFromFrameEvents(frameFunction);
}

void BlockTrap::resetBlackHole() {
	entered = false;
	GameLoop::unregisterFromFrameEvents(frameFunction);
}

void BlockTrap::contactedBy(Ball& ball) {
	switch (type)
	{
	case Type::MINE:
		triggerMine();
		break;
	case Type::SPRING:
		triggerSpring();
		break;
	case Type::TRAPDOOR:
		triggerTrapdoor();
		break;
	case Type::BLACK_HOLE:
		break;
	default:
		break;
	}
}

void BlockTrap::triggeredBy(Ball& ball) {
	if (type == Type::BLACK_HOLE) {
		triggerBlackHole();
	}
}

void BlockTrap::untriggeredBy(Ball& ball) {
	if (type == Type::BLACK_HOLE) {
		untriggerBlackHole();
	}
}

void BlockTrap::frameEvent() {
	if (type == Type::BLACK_HOLE) {
		blackHoleAttract();
	}
}

void BlockTrap::singleShot(int duration, void(BlockTrap::*callback)()) {
	timer.stop();
	timer.setSingleShot(true);
	timer.setInterval(duration);
	timer.disconnect();
	connect(&timer, &QTimer::timeout, this, callback);
	timer.start();
}

void BlockTrap::reset() {
	timer.stop();
	switch (type)
	{
	case Type::MINE:
		resetMine();
		break;
	case Type::SPRING:
		resetSpring();
		break;
	case Type::TRAPDOOR:
		resetTrapdoor();
		break;
	case Type::BLACK_HOLE:
		resetBlackHole();
		break;
	default:
		break;
	}
}

BlockTrap::~BlockTrap(){
	if (type == Type::BLACK_HOLE && entered) {
		GameLoop::unregisterFromFrameEvents(frameFunction);
	}
}

void BlockTrap::BlockTrapdoorCallback::operator()(Node* node, NodeVisitor* nv) {
	if (!trap->open) {
		NodeCallback::operator()(node, nv);
	}
}