#include "Board.h"
#include "Teeter.h"
#include "Level.h"
#include "Block.h"
#include "BlockStandard.h"
#include "BlockCheckpoint.h"
#include "BlockButton.h"
#include "BlockGhost.h"
#include "BlockElevator.h"
#include "BlockTrap.h"
#include "BlockItem.h"
#include "BlockPortal.h"
#include "Ball.h"

Board::Board(Level& level) : GameObject(level) {}

void Board::build(QImage& image) {
	group = new osg::Group;
	board = new osg::Geode;
	getTeeter().PX::createCompound(getId(), PX::NO_GRAVITY);


	boardTransform = new MatrixTransform(osg::Matrix::identity());
	boardTransform->addChild(board);
	getTeeter().setTransform(getId(), boardTransform.get());

	group->addChild(boardTransform);
	group->getOrCreateStateSet()->addUniform(new Uniform("a_TextureOffset", Vec3()));
	boardShader = new ShaderProgram(group);


	if (!getTeeter().shaders.count("boardShader.vert")) {
		getTeeter().shaders["boardShader.vert"] = new ShaderProgram::Shader("data/Teeter/boardShader.vert", Shader::VERTEX);
		getTeeter().shaders["boardShader.vert"]->addCodeFileInjection("data/rotation.glsl", "//HOTSPOT_PRE_MAIN");
		getTeeter().shaders["boardShader.vert"]->addCodeFileInjection("data/phong_vert.glsl", "//HOTSPOT_PRE_MAIN");
	}
	boardShader->setVertexShader(getTeeter().shaders["boardShader.vert"]);
	if (!getTeeter().shaders.count("boardShader.frag")) {
		getTeeter().shaders["boardShader.frag"] = new ShaderProgram::Shader("data/Teeter/boardShader.frag", Shader::FRAGMENT);
		getTeeter().shaders["boardShader.frag"]->addCodeFileInjection("data/classicnoise3D.glsl", "//HOTSPOT_PRE_MAIN");
		getTeeter().shaders["boardShader.frag"]->addCodeFileInjection("data/wood.glsl", "//HOTSPOT_PRE_MAIN");
		getTeeter().shaders["boardShader.frag"]->addCodeFileInjection("data/phong_frag.glsl", "//HOTSPOT_PRE_MAIN");
	}
	boardShader->setFragmentShader(getTeeter().shaders["boardShader.frag"]);


	for (int y = 0; y < image.height(); ++y) {
		for (int x = 0; x < image.width(); ++x) {
			QColor color = QColor::fromRgba(image.pixel(x, y));
			qreal hf, sf, bf, af;
			color.getHsvF(&hf, &sf, &bf, &af);
			// +0.5 -> round to nearest
			int h = hf * 360 + 0.5;
			int s = sf * 100 + 0.5;
			int b = bf * 100 + 0.5;
			int a = af * 255 + 0.5;
			// std::cout << "Block " << x << "|" << y << " has hsbaF: " << hf << "," << sf << "," << bf << "," << af << std::endl;
			// std::cout << "Block " << x << "|" << y << " has hsba: " << h << "," << s << "," << b << "," << a << std::endl;

			if (b == 0) {
				continue;
			}

			if (!idGroups.count(a)) {
				idGroups[a] = std::vector<ref_ptr<Block>>();
			}

			std::function<Block*(Level& level, float x, float y, int s, int b, std::vector<ref_ptr<Block>>& group)> factory;

			if (BlockStandard::isHue(h))
				factory = BlockStandard::factory;
			else if (BlockCheckpoint::isHue(h))
				factory = BlockCheckpoint::factory;
			else if (BlockButton::isHue(h))
				factory = BlockButton::factory;
			else if (BlockGhost::isHue(h))
				factory = BlockGhost::factory;
			else if (BlockElevator::isHue(h))
				factory = BlockElevator::factory;
			else if (BlockTrap::isHue(h))
				factory = BlockTrap::factory;
			else if (BlockItem::isHue(h))
				factory = BlockItem::factory;
			else if (BlockPortal::isHue(h))
				factory = BlockPortal::factory;
			else
				continue;

			ref_ptr<Block> block = factory(getLevel(), x - image.width() / 2.0 + 0.5, -y + image.height() / 2.0 - 0.5, s, b, idGroups[a]);
			blocks.push_back(block.get());
			idGroups[a].push_back(block.get());
		}
	}

	getLevel().addChild(group);

	getTeeter().PX::finishCompound(getId(), 1.0f);
	getTeeter().PX::setMatrix(getId(), boardTransform->getMatrix());

	// PhysX spherical joint in board center
	int jointId = getTeeter().PX::getNextJointId();
	getTeeter().PX::createSphericalJoint(jointId, getId(), osg::Matrix::rotate(PI_2, osg::Vec3(0.0, 1.0, 0.0)), -1, osg::Matrix::rotate(PI_2, osg::Vec3(0.0, 1.0, 0.0)));
	getTeeter().PX::setSphericalJointLimit(jointId, PI_2 / 3.0, PI_2 / 3.0, 0.001f);
	getTeeter().PX::setMaxAngularVelocity(getId(), 1.0);

	// PhysX distance joint (=spring) in each corner
	jointId = getTeeter().PX::getNextJointId();
	getTeeter().PX::createDistanceJoint(jointId, getId(), osg::Matrix::translate(-8, -8, 0), -1, osg::Matrix::translate(-8, -8, 0));
	getTeeter().PX::setDistanceJointLimit(jointId, 0, 1, 10, 1);
	jointId = getTeeter().PX::getNextJointId();
	getTeeter().PX::createDistanceJoint(jointId, getId(), osg::Matrix::translate(8, -8, 0), -1, osg::Matrix::translate(8, -8, 0));
	getTeeter().PX::setDistanceJointLimit(jointId, 0, 1, 10, 1);
	jointId = getTeeter().PX::getNextJointId();
	getTeeter().PX::createDistanceJoint(jointId, getId(), osg::Matrix::translate(-8, 8, 0), -1, osg::Matrix::translate(-8, 8, 0));
	getTeeter().PX::setDistanceJointLimit(jointId, 0, 1, 10, 1);
	jointId = getTeeter().PX::getNextJointId();
	getTeeter().PX::createDistanceJoint(jointId, getId(), osg::Matrix::translate(8, 8, 0), -1, osg::Matrix::translate(8, 8, 0));
	getTeeter().PX::setDistanceJointLimit(jointId, 0, 1, 10, 1);

	getTeeter().PX::setAngularDamping(getId(), 1);

	synchronizerId = getLevel().getNextId();
	synchronizer = new Geode();
	ref_ptr<ShapeDrawable> syncDrawable = new ShapeDrawable(new Box(Vec3(), image.width(), image.height(), 2));
	syncDrawable->setColor(Vec4(0.0,9.0,0.0,0.3));
	synchronizer->addDrawable(syncDrawable);
	ref_ptr<osg::MatrixTransform> synchronizerTransform = new osg::MatrixTransform();
	synchronizerTransform->addChild(synchronizer);
	synchronizerTransform->setCullCallback(new SynchronizerCallback(&getTeeter()));
	synchronizerTransform->getOrCreateStateSet()->setMode(GL_BLEND, osg::StateAttribute::ON);
	synchronizerTransform->getOrCreateStateSet()->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	getLevel().addChild(synchronizerTransform);
	getTeeter().setTransform(synchronizerId, synchronizerTransform);
}

ShaderProgram::ShaderGroup* Board::getShaderGroup(std::string type) {
	if (!shaderGroups.count(type)) {
		shaderGroups[type] = new ShaderProgram::ShaderGroup();
		group->addChild(shaderGroups[type]);
	}
	return shaderGroups[type].get();
}

void Board::reset(bool hardreset) {
	getTeeter().PX::setMatrix(getId(), osg::Matrix::identity());
	if (hardreset) {
		for (std::vector<ref_ptr<Block>>::iterator i = blocks.begin(); i != blocks.end(); ++i) {
			(*i)->reset();
		}
	}
}

void Board::registerTriggerableBlock(Block* block, int actorId) {
	triggerableBlocks[actorId] = block;
}
void Board::registerContactableBlock(Block* block, int actorId) {
	contactableBlocks[actorId] = block;
}

void Board::onContact(const physx::PxContactPairHeader& pairHeader, const physx::PxContactPair* pairs, physx::PxU32 nbPairs) {
	float strength = 0.0;
	for (PxU32 i = 0; i < nbPairs; i++)
	{
		PxContactPairPoint* contacts = new PxContactPairPoint[pairs[i].contactCount];
		PxU32 nbContacts = pairs[i].extractContacts(contacts, pairs[i].requiredBufferSize);
		for (PxU32 j = 0; j < nbContacts; j++)
		{
			PxVec3 point = contacts[j].position;
			PxVec3 impulse = contacts[j].impulse;
			strength = impulse.magnitude();
		}
		delete[] contacts;
	}

	if (getTeeter().PX::areFilterGroups(Level::Groups::BALL, Level::Groups::BOARD, pairs->shapes)) {
		if (strength > 0.01) {
			// Choose a random number between 1 and 8
			const unsigned long n = 8;
			const unsigned long divisor = (RAND_MAX + 1) / n;

			unsigned long k;
			do { k = std::rand() / divisor; } while (k >= n);

			// Play that random sound
			getTeeter().FM::playSound("hit" + std::to_string(k + 1), strength * 2);
		}
	}
	
	for (std::unordered_map<int, ref_ptr<Block>>::iterator i = contactableBlocks.begin(); i != contactableBlocks.end(); ++i) {
		if (getTeeter().PX::areActors(getLevel().getBall()->getId(), (*i).first, pairHeader.actors)) {
			getLevel().getBall()->contact(*(*i).second.get());
		}
	}
}

void Board::onTrigger(const physx::PxTriggerPair& pair, physx::PxRigidActor* const actors[2]) {
	for (std::unordered_map<int, ref_ptr<Block>>::iterator i = triggerableBlocks.begin(); i != triggerableBlocks.end(); ++i) {
		if (getTeeter().PX::areActors(getLevel().getBall()->getId(), (*i).first, actors)) {
			if (pair.status & PxPairFlag::eNOTIFY_TOUCH_FOUND)
				getLevel().getBall()->trigger(*(*i).second.get());
			else if (pair.status & PxPairFlag::eNOTIFY_TOUCH_LOST)
				getLevel().getBall()->untrigger(*(*i).second.get());
		}
	}
}

void Board::SynchronizerCallback::operator()(Node* node, NodeVisitor* nv) {
	if (teeter->getState() == Teeter::LevelState::SYNCHRONIZING) {
		NodeCallback::operator()(node, nv);
	}
}