#include "Ribbon.h"
#include "Teeter.h"

Ribbon::Ribbon(osg::Vec3 colorRGB, osg::MatrixTransform* follow, int followPhysxId, Teeter& teeter, osg::MatrixTransform* parentTransform)
	: followPhysxId(followPhysxId)
	, teeter(teeter)
{
	// create the ribbon
	ribbon = createRibbon(colorRGB);
	parent = parentTransform;

	// create the geode for it
	osg::ref_ptr<osg::Geode> geode = new osg::Geode;
	ribbonShader = new ShaderProgram(geode);
	ribbonShader->setVertexShader(teeter.shaders["standardShader.vert"]);
	ribbonShader->setFragmentShader(teeter.shaders["standardShader.frag"]);
	geode->addDrawable(ribbon);
	geode->getOrCreateStateSet()->setMode(GL_BLEND, osg::StateAttribute::ON);
	geode->getOrCreateStateSet()->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	// attach it to the parent node
	parent->addChild(geode.get());

	// add a callback for the node that should be followed
	follow->addUpdateCallback(this);
}

osg::Geometry* Ribbon::createRibbon(const osg::Vec3& colorRGB)
{
	// initialize arrays
	osg::ref_ptr<osg::Vec3Array> vertices = new osg::Vec3Array(g_numPoints);
	osg::ref_ptr<osg::Vec3Array> normals = new osg::Vec3Array(g_numPoints);
	osg::ref_ptr<osg::Vec4Array> colors = new osg::Vec4Array(g_numPoints);

	// set all points to the origin
	osg::Vec3 origin = osg::Vec3(0.0f, 0.0f, 0.0f);
	osg::Vec3 normal = osg::Vec3(0.0f, 0.0f, 1.0f);
	for (unsigned int i = 0; i<g_numPoints - 1; i += 2)
	{
		(*vertices)[i] = origin; (*vertices)[i + 1] = origin;
		(*normals)[i] = normal; (*normals)[i + 1] = normal;

		float alpha = sinf(osg::PI * (float)i / (float)g_numPoints);
		(*colors)[i] = osg::Vec4(colorRGB, alpha);
		(*colors)[i + 1] = osg::Vec4(colorRGB, alpha);
	}

	// set geometry attributes
	osg::ref_ptr<osg::Geometry> geom = new osg::Geometry;
	geom->setDataVariance(osg::Object::DYNAMIC);
	geom->setUseDisplayList(false);
	geom->setUseVertexBufferObjects(true);

	// bind the arrays to the geometry
	geom->setVertexArray(vertices.get());
	geom->setNormalArray(normals.get(), osg::Array::Binding::BIND_PER_VERTEX);
	geom->setColorArray(colors.get(), osg::Array::Binding::BIND_PER_VERTEX);
	geom->addPrimitiveSet(new osg::DrawArrays(GL_QUAD_STRIP, 0, g_numPoints));

	return geom.release();
}

void Ribbon::operator()(osg::Node* node, osg::NodeVisitor* nv)
{
	osg::MatrixTransform* trans = static_cast<osg::MatrixTransform*>(node);
	if (trans)
	{
		osg::Vec3Array* vertices = static_cast<osg::Vec3Array*>(ribbon->getVertexArray());
		osg::Vec3Array* normals = static_cast<osg::Vec3Array*>(ribbon->getNormalArray());

		// move all points forward by two (freeing space for the two new points)
		for (unsigned int i = 0; i<g_numPoints - 3; i += 2)
		{
			(*vertices)[i] = (*vertices)[i + 2];
			(*vertices)[i + 1] = (*vertices)[i + 3];
			(*normals)[i] = (*normals)[i + 2];
			(*normals)[i + 1] = (*normals)[i + 3];
		}

		// Convert the matrix to the board's transform, and only use the translation
		osg::Matrix transMatrix = osg::Matrix::translate((trans->getMatrix() * (parent->getInverseMatrix())).getTrans());
		// Use the rotation from the current velocity of the ball
		osg::Vec3 linVel = teeter.getLinearVelocity(followPhysxId);
		double angle = atan2(1 - linVel.x(), - linVel.y());
		osg::Matrix rotMatrix = osg::Matrix::rotate(angle, osg::Vec3(0, 0, 1));
		// create 2 new points based on the matrix
		(*vertices)[g_numPoints - 2] = osg::Vec3(-g_halfWidth, 0.0f, 0.0f) * rotMatrix * transMatrix;
		(*vertices)[g_numPoints - 1] = osg::Vec3(g_halfWidth, 0.0f, 0.0f) * rotMatrix * transMatrix;
		vertices->dirty();

		// set the corresponding normals
		osg::Vec3 normal = osg::Vec3(0.0f, 0.0f, 1.0f) * rotMatrix * transMatrix;
		normal.normalize();
		(*normals)[g_numPoints - 2] = normal;
		(*normals)[g_numPoints - 1] = normal;
		normals->dirty();

		ribbon->dirtyBound();
	}
	traverse(node, nv);
}