#include "Teeter.h"
#include "Level.h"
#include "Ground.h"
#include "Board.h"
#include "Ball.h"

Teeter::Teeter(OsgViewerWidget* osgViewerWidget) {
	GameLoop::lock();

	// Prepare camera
	osg::Camera* camera = getCamera();
	camera->setGraphicsContext(osgViewerWidget->getGraphicsContext());
	camera->setClearColor(osg::Vec4(0.2, 0.2, 0.6, 1.0));
	camera->setViewport(new osg::Viewport(0, 0, 800, 600));
	camera->setProjectionMatrixAsPerspective(30.0f, static_cast<double>(800) / static_cast<double>(600), 1.0f, 10000.0f);

	createScene();
	loadLevel(0);
	
	// Let the show begin!
	setSceneData(osgScene.get());

	handCamera = new HandCamera();
	handCamera->setHeight(30);
	handCamera->setDistance(25);
	handCamera->setAngle(osg::PI);
	handCamera->setCenter(osg::Vec3(0, -2, 0));
	handCamera->setVerticalAxisFixed(true);
	setCameraManipulator(handCamera.get());

	// Add Leap support
	LeapManager::setListener(*this);
	// Add keyboard support
	osgViewerWidget->registerKeyListener(this);
	// Add all of the wood sounds
	for (int i = 1; i <= 8; i++)
		FM::addSound("hit" + std::to_string(i), "data/Teeter/wood-ball-hit-" + std::to_string(i) + ".mp3");

	// Add menu hud

	MenuItemFolder* mainMenu = new MenuItemFolder("Menu");
	mainMenu->addItem(new MenuItemAction([&]() { requested = RequestType::RESTART; state = LevelState::SYNCHRONIZING; }, "Restart"));
	mainMenu->addItem(new MenuItemAction([&]() { requested = RequestType::NEXT; state = LevelState::SYNCHRONIZING; }, "Next Level"));
	mainMenu->addItem(new MenuItemAction([&]() { requested = RequestType::PREVIOUS; state = LevelState::SYNCHRONIZING; }, "Previous Level"));
	mainMenu->addItem(new MenuItemAction([&]() { pauseMenu->hide(); state = LevelState::SYNCHRONIZING; }, "Continue"));

	pauseMenu = new MenuHUD();
	pauseMenu->attachToScene(osgScene, this);
	pauseMenu->setMainMenu(mainMenu);
	pauseMenu->setCurrentMenu(mainMenu);

	mainMenu = new MenuItemFolder("Menu");
	mainMenu->addItem(new MenuItemAction([&]() { requested = RequestType::RESTART; state = LevelState::SYNCHRONIZING; }, "Restart"));
	mainMenu->addItem(new MenuItemAction([&]() { requested = RequestType::NEXT; state = LevelState::SYNCHRONIZING; }, "Next Level"));
	mainMenu->addItem(new MenuItemAction([&]() { requested = RequestType::PREVIOUS; state = LevelState::SYNCHRONIZING; }, "Previous Level"));

	endMenu = new MenuHUD();
	endMenu->attachToScene(osgScene, this);
	endMenu->setMainMenu(mainMenu);
	endMenu->setCurrentMenu(mainMenu);

	notification = new TextHUD();
	notification->setRelativePosition(Vec2(0.5,0.05));
	notification->setAlignment(osgText::TextBase::AlignmentType::CENTER_BOTTOM);
	notification->attachToScene(osgScene, this);

	state = LevelState::SYNCHRONIZING;
	GameLoop::unlock();
}

void Teeter::onConnect(const Leap::Controller& controller) {
	controller.enableGesture(Leap::Gesture::TYPE_CIRCLE);
	controller.enableGesture(Leap::Gesture::TYPE_SCREEN_TAP);
}

int Teeter::fps() {
	return 50;
}

float Teeter::simulationStep() {
	return 1 / 30.f;
}

void Teeter::start() {
	state = LevelState::START;
}

void Teeter::won() {
	state = LevelState::WON;
	if (!notification->isVisible()) {
		notification->setText("Well Done!");
		notification->setColor(Vec4(0.0, 0.9, 0.0, 1.0));
		notification->show();
	}
}

void Teeter::lost() {
	state = LevelState::LOST;
	if (!notification->isVisible()) {
		notification->setText("Not even close...");
		notification->setColor(Vec4(0.9, 0.0, 0.0, 1.0));
		notification->show();
	}
}

void Teeter::end(const Controller& controller) {
	GameLoop::lock();
	// Check for menu
	if (endMenu->isVisible()) {
		endMenu->onFrame(controller);
		GameLoop::unlock();
		return;
	}
	else {
		endMenu->show();
	}
	GameLoop::unlock();
}

void Teeter::pause(const Controller& controller) {
	GameLoop::lock();
	// Check for menu
	if (pauseMenu->isVisible()) {
		pauseMenu->onFrame(controller);
		GameLoop::unlock();
		return;
	}
	else {
		pauseMenu->show();
	}
	GameLoop::unlock();
}

void Teeter::synchronize(Leap::Frame& frame) {
	// show synchronize hint
	state = LevelState::SYNCHRONIZING;

	GameLoop::lock();
	if (!notification->isVisible()) {
		notification->setText("Bring your hand in rest position.");
		notification->setColor(Vec4(0.0, 0.9, 0.0, 1.0));
		notification->show();
	}

	if (!frame.hands().isEmpty()) {
		// Get the first hand
		Leap::Hand hand = frame.hands()[0];
		Leap::Vector normal = hand.palmNormal();
		Leap::Vector direction = hand.direction();

		osg::Matrix rotation = osg::Matrix::rotate(osg::PI + handCamera->getAngle(), Vec3(0.0, 0.0, 1.0));

		Vec3 rot = getRotationInDirection(*transforms[level->getBoard()->getId()], Vec3(0.0, 0.0, -1.0), rotation * toOsgVec3(normal), osg::Vec3(0.0, 1.0, 0.0), toOsgVec3(direction),0);
		if (rot.length() < 0.25) {
			state = Teeter::LevelState::RUNNING;
			notification->hide();
		}
		else {
			transforms[level->getBoard()->getSynchronizerId()]->setMatrix(osg::Matrix::rotate(Vec3(0, 0, -1), rotation * toOsgVec3(normal)));
		}
	}
	GameLoop::unlock();
}

void Teeter::onFrame(const Controller& controller) {
	Leap::Frame frame = controller.frame();
	switch (state)
	{
	case Teeter::LevelState::START:
	case Teeter::LevelState::SYNCHRONIZING:
		if (handCamera->handleLeapFrame(frame)) {
			state = LevelState::CAMERA;
		}
		else {
			synchronize(frame);
		}
		break;
	case Teeter::LevelState::RUNNING:
		if (handCamera->handleLeapFrame(frame)) {
			state = LevelState::CAMERA;
			return;
		}
		else if(!openPauseMenu(frame)) {
			rotateBoard(frame);
		}
		break;
	case Teeter::LevelState::CAMERA:
		if (!handCamera->handleLeapFrame(frame)) {
			state = LevelState::SYNCHRONIZING;
			return;
		}
		break;
	case Teeter::LevelState::PAUSE:
		pause(controller);
		break;
	case Teeter::LevelState::WON:
	case Teeter::LevelState::LOST:
		end(controller);
	default:
		break;
	}
}

bool Teeter::openPauseMenu(Leap::Frame& frame) {
	for (int g = 0; g < frame.gestures().count(); ++g) {
		Leap::Gesture gesture = frame.gestures()[g];
		// Check for screen tap gesture
		if (gesture.type() == Leap::Gesture::TYPE_SCREEN_TAP) {
			state = LevelState::PAUSE;
			return true;
		}
	}
	return false;
}

void Teeter::rotateBoard(Leap::Frame& frame) {
	if (!frame.hands().isEmpty()) {
		// Get the first hand
		Leap::Hand hand = frame.hands()[0];
		Leap::Vector normal = hand.palmNormal();
		Leap::Vector direction = hand.direction();

		osg::Matrix rotation = osg::Matrix::rotate(osg::PI + handCamera->getAngle(), Vec3(0.0, 0.0, 1.0));

		GameLoop::lock();
		PX::setAngularVelocity(level->getBoard()->getId(),
			getRotationInDirection(*transforms[level->getBoard()->getId()], Vec3(0.0, 0.0, -1.0), rotation * toOsgVec3(normal), osg::Vec3(0.0, 1.0, 0.0), toOsgVec3(direction)) * 2);
		GameLoop::unlock();
	}
}

void Teeter::keyPressEvent(QKeyEvent* event) {
	switch (event->key()) {
	case Qt::Key_Up:
		keyUpPressed = true;
		break;
	case Qt::Key_Down:
		keyDownPressed = true;
		break;
	case Qt::Key_Left:
		keyLeftPressed = true;
		break;
	case Qt::Key_Right:
		keyRightPressed = true;
		break;
	default:
			return;
	}
	event->accept();
};

void Teeter::keyReleaseEvent(QKeyEvent* event) {
	if (pauseMenu->isVisible()) {
		pauseMenu->keyReleaseEvent(event);

		// If menu was closed
		if (!pauseMenu->isVisible())
			state = LevelState::SYNCHRONIZING;
		
		return;
	} else if (endMenu->isVisible()) {
		endMenu->keyReleaseEvent(event);
		return;
	}

	int currentLevel = level->getLevel();
	switch (event->key()) {
	case Qt::Key_Escape:
		if (!endMenu->isVisible()) {
			if (!pauseMenu->isVisible()) {
				state = LevelState::PAUSE;
			}
			else {
				state = LevelState::SYNCHRONIZING;
			}
			pauseMenu->toggleVisibility();
		}
		break;
	case Qt::Key_Up:
		keyUpPressed = false;
		break;
	case Qt::Key_Down:
		keyDownPressed = false;
		break;
	case Qt::Key_Left:
		keyLeftPressed = false;
		break;
	case Qt::Key_Right:
		keyRightPressed = false;
		break;
	case Qt::Key_R:
		loadLevel(currentLevel);
		break;
	case Qt::Key_Plus:
		loadLevel(currentLevel + 1);
		break;
	case Qt::Key_Minus:
		loadLevel(currentLevel - 1);
		break;
	default:
		return;
	}
	event->accept();
}


void Teeter::frameEvent() {
	GameLoop::lock();
	// Check for level switch event

	if (requested != NONE) {
		switch (requested) {
		case RequestType::RESTART:
			loadLevel(level->getLevel());
			break;
		case RequestType::NEXT:
			loadLevel(level->getLevel() + 1);
			break;
		case RequestType::PREVIOUS:
			loadLevel(level->getLevel() - 1);
			break;
		default:
			break;
		}
		requested = RequestType::NONE;
	}


	osg::Matrix rotation = osg::Matrix::identity();
	rotation.makeRotate(handCamera->getAngle(), Vec3(0.0, 0.0, 1.0));
	if (keyUpPressed)
		PX::addTorque(level->getBoard()->getId(), rotation * osg::Vec3(0.1, 0.0, 0.0), PxForceMode::eVELOCITY_CHANGE);
	if (keyDownPressed)
		PX::addTorque(level->getBoard()->getId(), rotation * osg::Vec3(-0.1, 0.0, 0.0), PxForceMode::eVELOCITY_CHANGE);
	if (keyLeftPressed)
		PX::addTorque(level->getBoard()->getId(), rotation * osg::Vec3(0.0, 0.1, 0.0), PxForceMode::eVELOCITY_CHANGE);
	if (keyRightPressed)
		PX::addTorque(level->getBoard()->getId(), rotation * osg::Vec3(0.0, -0.1, 0.0), PxForceMode::eVELOCITY_CHANGE);
	GameLoop::unlock();
}

void Teeter::onContact(const physx::PxContactPairHeader& pairHeader, const physx::PxContactPair* pairs, physx::PxU32 nbPairs) {
	if (PX::areActors(level->getGround()->getId(), level->getBall()->getId(), pairHeader.actors)) {
		level->getBall()->contact(*level->getGround());
	}
	else if (PX::containsActor(level->getBall()->getId(), pairHeader.actors) && PX::areFilterGroups(Level::Groups::BOARD, Level::Groups::BALL, pairs->shapes)) {
		level->getBoard()->onContact(pairHeader, pairs, nbPairs);
	}
}

void Teeter::onTrigger(PxTriggerPair* pairs, PxU32 count)
{
	for (PxU32 i = 0; i < count; i++)
	{
		// ignore pairs when shapes have been deleted
		if (pairs[i].flags & (PxTriggerPairFlag::eDELETED_SHAPE_TRIGGER | PxTriggerPairFlag::eDELETED_SHAPE_OTHER))
			continue;
		PxShape* shapes[2];
		shapes[0] = pairs[i].triggerShape;
		shapes[1] = pairs[i].otherShape;

		PxRigidActor* actors[2];
		actors[0] = pairs[i].triggerActor;
		actors[1] = pairs[i].otherActor;
		
		if (PX::containsActor(level->getBall()->getId(), actors) && PX::areFilterGroups(Level::Groups::BOARD, Level::Groups::BALL, shapes)) {
			level->getBoard()->onTrigger(pairs[i], actors);
		}
	}
}

void Teeter::createScene() {
	GameLoop::lock();
	// Prepare world
	PX::createScene(osg::Vec3(0.0f, 0.0f, -9.81f));
	osgScene = new Group();
	osgScene->getOrCreateStateSet()->addUniform(new Uniform("a_LightPos", Vec3f(8, -8, 16)));
	GameLoop::unlock();
}

void Teeter::loadLevel(int levelNumber) {
	GameLoop::lock();
	level = nullptr;
	level = new Level(levelNumber, osgScene, *this);
	GameLoop::unlock();
	state = LevelState::SYNCHRONIZING;
}

void Teeter::setTransform(int id, osg::MatrixTransform* transform) {
	transforms[id] = transform;
}

void Teeter::removeTransform(int id) {
	transforms.erase(id);
}