#include "AbstractHUD.h"

#include "ContainerHUD.h"


void AbstractHUD::attachToScene(osg::Group* node, osgViewer::View* view) {
	_geode = new osg::Geode();

	osg::StateSet* HUDStateSet = new osg::StateSet();
	_geode->setStateSet(HUDStateSet);

	// For this state set, turn blending on (so alpha texture looks right)
	HUDStateSet->setMode(GL_BLEND, osg::StateAttribute::ON);

	// Disable depth testing so geometry is draw regardless of depth values
	// of geometry already draw.
	HUDStateSet->setMode(GL_DEPTH_TEST, osg::StateAttribute::OFF);
	HUDStateSet->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	// Need to make sure this geometry is draw last. RenderBins are handled
	// in numerical order so set bin number to 11
	HUDStateSet->setRenderBinDetails(11, "RenderBin");

	// Projection node for defining view frustrum for HUD:
	HUDProjectionMatrix = new osg::Projection();

	isInContainer = false;

	// Initially set resolution to that of the viewport
	setResolution(view->getCamera()->getViewport()->width(), view->getCamera()->getViewport()->height());

	// For the HUD model view matrix use an identity matrix:
	HUDModelViewMatrix->setMatrix(osg::Matrix::identity());

	// Make sure the model view matrix is not affected by any transforms
	// above it in the scene graph:
	HUDModelViewMatrix->setReferenceFrame(osg::Transform::ABSOLUTE_RF_INHERIT_VIEWPOINT);

	HUDProjectionMatrix->addChild(HUDModelViewMatrix);
	HUDModelViewMatrix->addChild(_geode);

	_node = node;
	_view = view;

}

void AbstractHUD::attachToContainer(ContainerHUD* container)
{
	_container = container;
	isInContainer = true;
	container->addChild(this);
}

void AbstractHUD::setGeode(osg::Geode* geode) {
	_geode = geode;
}

osg::Geode* AbstractHUD::getGeode() {
	return _geode.get();
}

bool AbstractHUD::handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter&, osg::Object*, osg::NodeVisitor*) {
	// Refresh resolution on resizing event
	if (_width == 0 || ea.getEventType() == osgGA::GUIEventAdapter::RESIZE) {
		setResolution(ea.getWindowWidth(), ea.getWindowHeight());
	}
	return false;
}

void AbstractHUD::show() {
	if (visible)
		return;

	visible = true;

	if (!isInContainer) {
		// Add self to the corresponding node and reset to the main menu
		_node->addChild(HUDProjectionMatrix);

		// Initially set resolution to that of the viewport
		setResolution(_view->getCamera()->getViewport()->width(), _view->getCamera()->getViewport()->height());

		// Add listener for RESIZE events
		_view->addEventHandler(this);
	}
	else
		_container->showHUD(this);
	// Redraw the HUD
	redraw();
}

void AbstractHUD::hide() {
	if (!visible)
		return;
	
	visible = false;

	if (!isInContainer) {
		_node->removeChild(HUDProjectionMatrix);
		_view->removeEventHandler(this);
	}
	else
		_container->hideHUD(this);
}

void AbstractHUD::toggleVisibility() {
	if (visible)
		hide();
	else
		show();
}

void AbstractHUD::redraw() {
	if (!visible || _geode == nullptr)
		return;
	clear();
	draw();
}

void AbstractHUD::clear() {
	_geode->removeDrawables(0, _geode->getNumDrawables());
}

bool AbstractHUD::isVisible() {
	return visible;
}

int AbstractHUD::getWidth() {
	return _width;
}

int AbstractHUD::getHeight() {
	return _height;
}

void AbstractHUD::setResolution(int width, int height) {
	// Refresh the projection matrix according to the new resolution
	_width = width;
	_height = height;

	if (!isInContainer)
		HUDProjectionMatrix->setMatrix(osg::Matrix::ortho2D(0, width, 0, height));

	redraw();
}