#include "MenuHUD.h"

#include "TextHUD.h"
#include "ImageHUD.h"

MenuHUD::MenuHUD() 
: ContainerHUD()
{
	
}

AbstractMenuItem::AbstractMenuItem(char* text, char* image)
: _text(text)
, _image(image)
{
}

char* AbstractMenuItem::getText() {
	return _text;
}

char* AbstractMenuItem::getImage() {
	return _image;
}

MenuItemAction::MenuItemAction(std::function<void()> f, char* text, char* image)
: AbstractMenuItem(text, image)
, _f(f)
{
}

void MenuItemAction::select(MenuHUD* hud) {
	hud->hide();
	_f();
}

MenuItemFolder::MenuItemFolder(char* text, char* image)
: AbstractMenuItem(text, image)
{
}

void MenuItemFolder::selectNext() {
	highlighted = (highlighted + 1) % counter;
	_hud->redraw();
}

void MenuItemFolder::selectPrevious() {
	highlighted = (highlighted + counter - 1) % counter;
	_hud->redraw();
}

void MenuItemFolder::select(MenuHUD* hud) {
	// Set the menu and redraw
	_hud = hud;
	_hud->setCurrentMenu(this);
	_hud->draw();
}

void MenuItemFolder::callSelected() {
	// call the function of the selected item
	subItems.at(highlighted)->select(_hud);
}


void MenuItemFolder::addItem(AbstractMenuItem* subItem) {
	// Set self as parent folder and add it to subItems list
	subItems.push_back(subItem);
	counter++;
}

void MenuItemFolder::onFrame(const Controller& controller) {
	const Leap::Frame frame = controller.frame();
	const Leap::GestureList gestures = frame.gestures();
	for (int g = 0; g < gestures.count(); ++g) {
		Leap::Gesture gesture = gestures[g];

		// Check for circle gesture
		if (gesture.type() == Leap::Gesture::TYPE_CIRCLE) {
			Leap::CircleGesture circle = gesture;
			bool clockwise = (circle.pointable().direction().angleTo(circle.normal()) <= Leap::PI / 2);

			if (circle.state() == Leap::Gesture::STATE_STOP && circle.progress() > 0.9) {
				if (clockwise)
					selectNext();
				else
					selectPrevious();
			}
		}
		// Check for screen tap gesture
		else if (gesture.type() == Leap::Gesture::TYPE_SCREEN_TAP) {
			callSelected();
		}
	}
}

void MenuItemFolder::keyReleaseEvent(QKeyEvent* event) {
	switch (event->key()) {
	case Qt::Key_Down:
		selectNext();
		break;
	case Qt::Key_Up:
		selectPrevious();
		break;
	case Qt::Key_Return:
		callSelected();
		break;
	case Qt::Key_Escape:
		_hud->hide();
		break;
	default:
		return;
	}
	event->accept();
}

void MenuHUD::onFrame(Controller controller) {
	// Forward the leap signal to the folder
	_currentMenu->onFrame(controller);
}

void MenuHUD::keyReleaseEvent(QKeyEvent* event) {
	// Forward the key event to the folder
	_currentMenu->keyReleaseEvent(event);
}


void MenuHUD::show() {
	ContainerHUD::show();
	GameLoop::pause();
	_mainMenu->select(this);
}

void MenuHUD::hide() {
	ContainerHUD::hide();
	GameLoop::unpause();
}

void MenuHUD::setCurrentMenu(MenuItemFolder* menu) {
	_currentMenu = menu;
}

void MenuHUD::setMainMenu(MenuItemFolder* menu) {
	_mainMenu = menu;
}


void MenuHUD::draw() {
	if (_currentMenu != nullptr){
		clear();
		deleteHUDs();
		
		// Add text of this menu
		osg::ref_ptr<TextHUD> titleHUD = new TextHUD();
		titleHUD->setText(_currentMenu->getText());
		titleHUD->setSize(50);
		titleHUD->setRelativePosition(osg::Vec2(0.5f, 0.9f));
		titleHUD->setAlignment(osgText::TextBase::AlignmentType::CENTER_TOP);
		titleHUD->attachToContainer(this);

		// Add text of subitems
		for (int i = 0; i < _currentMenu->counter; i++) {
			osg::ref_ptr<TextHUD> textHUD = new TextHUD();
			textHUD->setText(_currentMenu->subItems.at(i)->getText());
			textHUD->setSize(40);
			if (i == _currentMenu->highlighted)
				textHUD->setSize(50);
			textHUD->setRelativePosition(osg::Vec2(0.5f, 0.6f - 0.1f * i));
			textHUD->attachToContainer(this);
		}
	}
}