#include "OsgFunctions.h"

/// <summary>
/// Gets the rotation (torque) to rotate the object with the currentPose along its orientation into direction.
/// The rotation is damped, angles below 0.1 won't result in a rotation.
/// </summary>
/// <param name="currentPose">The current pose.</param>
/// <param name="orientation">The orientation.</param>
/// <param name="direction">The direction.</param>
/// <param name="threshold">The threshold./param>
/// <returns>The rotation vector</returns>
osg::Vec3 getRotationInDirection(osg::MatrixTransform& currentPose, osg::Vec3 orientation, osg::Vec3 direction, float threshold) {
	osg::Vec3 old = currentPose.getMatrix().getRotate() * orientation;
	osg::Vec3f axis;
	osg::Quat::value_type angle;
	osg::Quat rot;
	rot.makeRotate(old, direction);
	rot.getRotate(angle, axis);
	angle = angle > threshold ? angle : 0;
	return axis * angle;
}

/// <summary>
/// Gets the rotation (torque) to rotate the object with the currentPose along its orientationU into directionU and along its orientationV into directionV.
/// The rotation is damped, angles below 0.1 won't result in a rotation.
/// Behaviour is undefined if there is no possible torque to rotate the orientation vectors into the direction vectors.
/// </summary>
/// <param name="currentMatrix">The current matrix.</param>
/// <param name="orientationU">The orientation u vector.</param>
/// <param name="directionU">The direction u vector.</param>
/// <param name="orientationV">The orientation v vector.</param>
/// <param name="directionV">The direction v vector.</param>
/// <param name="threshold">The threshold./param>
/// <returns>The rotation vector</returns>
osg::Vec3 getRotationInDirection(const osg::Matrix& currentMatrix, osg::Vec3 orientationU, osg::Vec3 directionU, osg::Vec3 orientationV, osg::Vec3 directionV, float threshold) {
	osg::Vec3 u0 = currentMatrix.getRotate() * orientationU;
	osg::Vec3 v0 = currentMatrix.getRotate() * orientationV;

	osg::Vec3 u1 = u0;
	osg::Vec3 u2 = directionU;

	osg::Quat q2;
	q2.makeRotate(u1, u2);

	osg::Vec3 v2 = directionV;
	osg::Vec3 v1 = q2.conj() * v2;

	osg::Plane u0_plane = osg::Plane(u0, osg::Vec3());

	osg::Vec3 v0_proj = project(v0, u0_plane);
	osg::Vec3 v1_proj = project(v1, u0_plane);

	osg::Quat q1;
	q1.makeRotate(v0_proj, v1_proj);

	osg::Quat q = q2 * q1;

	osg::Vec3f axis;
	osg::Quat::value_type angle;
	q.getRotate(angle, axis);
	angle = angle > threshold ? angle : 0;
	return axis * angle;
}

/// <summary>
/// Gets the rotation (torque) to rotate the object with the currentPose along its orientationU into directionU and along its orientationV into directionV.
/// The rotation is damped, angles below 0.1 won't result in a rotation.
/// Behaviour is undefined if there is no possible torque to rotate the orientation vectors into the direction vectors.
/// </summary>
/// <param name="currentPose">The current pose.</param>
/// <param name="orientationU">The orientation u vector.</param>
/// <param name="directionU">The direction u vector.</param>
/// <param name="orientationV">The orientation v vector.</param>
/// <param name="directionV">The direction v vector.</param>
/// <param name="threshold">The threshold.</param>
/// <returns>The rotation vector</returns>
osg::Vec3 getRotationInDirection(osg::MatrixTransform& currentPose, osg::Vec3 orientationU, osg::Vec3 directionU, osg::Vec3 orientationV, osg::Vec3 directionV, float threshold) {
	return getRotationInDirection(currentPose.getMatrix(), orientationU, directionU, orientationV, directionV, threshold);
}

/// <summary>
/// Projects the specified vector on the specified plane.
/// </summary>
/// <param name="vector">The vector.</param>
/// <param name="plane">The plane.</param>
/// <returns>the projected vector</returns>
osg::Vec3 project(osg::Vec3 vector, osg::Plane plane) {
	osg::Vec3 normal = plane.getNormal();
	normal.normalize();

	double distance = plane.distance(vector);
	return vector - normal*distance;
}