#include "ShaderProgram.h"

#include <QFile>

ShaderProgram::CodeInjection::CodeInjection(std::string codeText, std::string hotspotname) {
	hotspot = QString::fromStdString(hotspotname);
	code = QString::fromStdString(codeText);
}

void ShaderProgram::CodeInjection::addToShader(Shader* shader) {
	shaders.insert(shader);
}

void ShaderProgram::CodeInjection::removeFromShader(Shader* shader) {
	shaders.erase(shader);
}


void ShaderProgram::CodeInjection::setCode(std::string codeText) {
	setCode(QString::fromStdString(codeText));
}
void ShaderProgram::CodeInjection::setCode(QString codeText, bool notify) {
	code = codeText;
	if (notify) {
		std::cout << "CodeInjection changed" << std::endl;
		for (std::unordered_set<Shader*>::iterator i = shaders.begin(); i != shaders.end(); ++i) {
			(*i)->notifyChanged("");
		}
	}
}

ShaderProgram::CodeFileInjection::CodeFileInjection(const std::string filename, std::string hotspotname)
: CodeInjection("", hotspotname)
, filename(filename)
{
	setCode(readFile(filename), false);
}

void ShaderProgram::CodeFileInjection::notifyChanged(const std::string filename) {
	setCode(readFile(filename));
}

void ShaderProgram::CodeFileInjection::addToShader(Shader* shader) {
	if (shaders.size() == 0)
		watchOn(filename);
	CodeInjection::addToShader(shader);
}

void ShaderProgram::CodeFileInjection::removeFromShader(Shader* shader) {
	CodeInjection::removeFromShader(shader);
	if (shaders.size() == 0)
		stopWatching(filename);
}

ShaderProgram::CodeFileInjection::~CodeFileInjection() {
	stopWatching(filename);
}


ShaderProgram::Shader::Shader(const std::string filename, osg::Shader::Type type) {
	shader = new osg::Shader(type);
	shader->loadShaderSourceFromFile(filename);
	shader->setName(filename);
	watchOn(filename);
}

void ShaderProgram::Shader::addCodeInjection(CodeInjection* injection) {
	injections.push_back(injection);
	injection->addToShader(this);
	rebuild();
}

void ShaderProgram::Shader::addCodeFileInjection(const std::string filename, const std::string hotspot) {
	addCodeInjection(new CodeFileInjection(filename, hotspot));
}

void ShaderProgram::Shader::addCodeTextInjection(const std::string code, const std::string hotspot) {
	addCodeInjection(new CodeInjection(code, hotspot));
}

void ShaderProgram::Shader::removeCodeInjection(CodeInjection* injection) {
	injection->removeFromShader(this);
	injections.remove(injection);
	rebuild();
}

void ShaderProgram::Shader::notifyChanged(const std::string filename) {
	std::cout << shader->getTypename() << " Shader " << shader->getName() << " changed" << std::endl;
	rebuild();
}

ShaderProgram::Shader::~Shader() {
	for (std::list<osg::ref_ptr<CodeInjection>>::iterator i = injections.begin(); i != injections.end(); ++i) {
		(*i)->removeFromShader(this);
	}
	stopWatching(shader->getFileName());
}

void ShaderProgram::Shader::rebuild() {
	if (injections.empty()) {
		shader->loadShaderSourceFromFile(shader->getFileName());
	}
	else {
		QString mainCode = readFile(shader->getFileName());
		for (std::list<osg::ref_ptr<CodeInjection>>::iterator i = injections.begin(); i != injections.end(); ++i) {
			mainCode.replace((*i)->getHotspot(), (*i)->getCode() + "\n" + (*i)->getHotspot());
		}
		shader->setShaderSource(mainCode.toStdString());
	}
}




ShaderProgram::ShaderProgram(osg::Node* node) : node(node) {
	program = new osg::Program;
	node->getOrCreateStateSet()->setAttributeAndModes(program.get(), osg::StateAttribute::ON);
}

void ShaderProgram::setVertexShader(Shader* shader) {
	removeVertexShader();
	vertexShader = shader;
	program->addShader(shader->getShader());
}

void ShaderProgram::setVertexShader(const std::string filename) {
	setVertexShader(new Shader(filename, osg::Shader::VERTEX));
}

void ShaderProgram::removeVertexShader() {
	if (vertexShader) {
		program->removeShader(vertexShader->getShader());
		vertexShader = nullptr;
	}
}

void ShaderProgram::setGeometryShader(Shader* shader) {
	removeGeometryShader();
	geometryShader = shader;
	program->addShader(shader->getShader());
}

void ShaderProgram::setGeometryShader(const std::string filename) {
	setGeometryShader(new Shader(filename, osg::Shader::GEOMETRY));
}

void ShaderProgram::removeGeometryShader() {
	if (geometryShader) {
		program->removeShader(geometryShader->getShader());
		geometryShader = nullptr;
	}
}

void ShaderProgram::setFragmentShader(Shader* shader) {
	removeFragmentShader();
	fragmentShader = shader;
	program->addShader(shader->getShader());
}

void ShaderProgram::setFragmentShader(const std::string filename) {
	setFragmentShader(new Shader(filename, osg::Shader::FRAGMENT));
}

void ShaderProgram::removeFragmentShader() {
	if (fragmentShader) {
		program->removeShader(fragmentShader->getShader());
		fragmentShader = nullptr;
	}
}

ShaderProgram::~ShaderProgram() {
	removeFragmentShader();
	removeGeometryShader();
	removeVertexShader();
	node->getOrCreateStateSet()->removeAttribute(program.get());
}

QString ShaderProgram::readFile(std::string filename) {
	QFile file(QString::fromStdString(filename));
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		return "";
	}

	QByteArray total;
	QByteArray line;
	while (!file.atEnd()) {
		line = file.read(1024);
		total.append(line);
	}

	return QString(total);
}