#pragma once

#include <list>
#include <unordered_set>

#include <osg/Geode>
#include <osg/Shader>
#include <osg/Program>

#include "FileAssociatedAsset.h"

/// <summary>
/// A shader program consisting of a fragment and a vertex shader that are
/// loaded from a file and kept in sync with that file.
/// There are some standard uniform and input bindings for vertex shaders:
/// in vec4 osg_Vertex
/// in vec3 osg_Normal
/// in vec4 osg_Color
/// in vec4 osg_SecondaryColor
/// in float osg_FogCoord
/// uniform mat4 osg_ModelViewMatrix
/// uniform mat4 osg_ModelViewProjectionMatrix
/// uniform mat4 osg_ProjectionMatrix
/// uniform mat3 osg_NormalMatrix
/// uniform mat4 osg_ViewMatrix
/// uniform mat4 osg_ViewMatrixInverse
/// uniform uint osg_FrameNumber
/// uniform float osg_FrameTime
/// uniform float osg_DeltaFrameTime
/// uniform float osg_SimulationTime
/// </summary>
class ShaderProgram : public osg::Referenced
{
public:
	class Shader;

	class CodeInjection : public virtual osg::Referenced{
	public:
		CodeInjection(std::string code, std::string hotspot);
		QString getHotspot() { return hotspot; }
		QString getCode() { return code; }
		void setCode(std::string codeText);
		void setCode(QString codeText, bool notify = true);
	friend class Shader;
	protected:
		virtual void addToShader(Shader* shader);
		virtual void removeFromShader(Shader* shader);
	protected:
		virtual ~CodeInjection() {};
		std::unordered_set<Shader*> shaders;
	private:
		QString code;
		QString hotspot;
	};

	class CodeFileInjection : public CodeInjection, public FileAssociatedAsset {
	public:
		CodeFileInjection(const std::string filename, std::string hotspot);
		std::string getFilename() { return filename; }
	protected:
		void notifyChanged(const std::string filename);
		virtual void addToShader(Shader* shader);
		virtual void removeFromShader(Shader* shader);
		virtual ~CodeFileInjection();
	private:
		std::string filename;
	};

	class Shader : public FileAssociatedAsset{
	public:
		Shader(const std::string filename, osg::Shader::Type type);
		void addCodeInjection(CodeInjection* injection);
		void addCodeFileInjection(const std::string filename, std::string hotspot);
		void addCodeTextInjection(const std::string code, std::string hotspot);
		void removeCodeInjection(CodeInjection* injection);
		osg::Shader* getShader() { return shader.get(); }
	friend class CodeInjection;
	protected:
		void notifyChanged(const std::string filename);
	protected:
		~Shader();
	private:
		void rebuild();
		osg::ref_ptr<osg::Shader> shader;
		std::list<osg::ref_ptr<CodeInjection>> injections;
	};

	class ShaderGroup : public osg::Group {
	public:
		osg::ref_ptr<ShaderProgram> shaderProgram;
	};

public:
	ShaderProgram(osg::Node* node);

	void setVertexShader(Shader* shader);
	void setVertexShader(const std::string filename);
	Shader* getVertexShader() { return vertexShader.get(); }
	void removeVertexShader();
	void setGeometryShader(Shader* shader);
	void setGeometryShader(const std::string filename);
	Shader* getGeometryShader() { return geometryShader.get(); }
	void removeGeometryShader();
	void setFragmentShader(Shader* shader);
	void setFragmentShader(const std::string filename);
	Shader* getFragmentShader() { return fragmentShader.get(); }
	void removeFragmentShader();

	static QString readFile(std::string filename);

protected:

	~ShaderProgram();

private:
	osg::ref_ptr<osg::Node> node;
	osg::ref_ptr<osg::Program> program;
	osg::ref_ptr<Shader> vertexShader;
	osg::ref_ptr<Shader> geometryShader;
	osg::ref_ptr<Shader> fragmentShader;
};