#include "PhysxScene.h"


/// <summary>
/// The filter shader used by PhysX collision detection
/// </summary>
/// <param name="attributes0">The first collision object's attributes.</param>
/// <param name="filterData0">The first collision object's data.</param>
/// <param name="attributes1">The second collision object's attributes.</param>
/// <param name="filterData1">The second collision object's data.</param>
/// <param name="pairFlags">The pair flags.</param>
/// <param name="constantBlock">The constant block.</param>
/// <param name="constantBlockSize">Size of the constant block.</param>
/// <returns>The PxFilterFlag that indicates how to react to the collision</returns>
PxFilterFlags TeetairFilterShader(
	PxFilterObjectAttributes attributes0, PxFilterData filterData0,
	PxFilterObjectAttributes attributes1, PxFilterData filterData1,
	PxPairFlags& pairFlags, const void* constantBlock, PxU32 constantBlockSize)
{
	// trigger pairs (A,B) where one object is a trigger and
	// the filtermask of A contains the ID of B and vice versa.
	if (PxFilterObjectIsTrigger(attributes0) || PxFilterObjectIsTrigger(attributes1))
	{
		if ((filterData0.word0 & filterData1.word1) && (filterData1.word0 & filterData0.word1))
			pairFlags = PxPairFlag::eTRIGGER_DEFAULT;
		return PxFilterFlag::eDEFAULT;
	}

	// generate contacts for all that were not filtered above
	pairFlags = PxPairFlag::eCONTACT_DEFAULT;

	// trigger the contact callback for pairs (A,B) where
	// the filtermask of A contains the ID of B and vice versa.
	if ((filterData0.word0 & filterData1.word1) && (filterData1.word0 & filterData0.word1))
		pairFlags |= PxPairFlag::eNOTIFY_TOUCH_FOUND | PxPairFlag::eNOTIFY_CONTACT_POINTS | PxPairFlag::eCCD_LINEAR;

	return PxFilterFlag::eDEFAULT;
}

bool PhysxScene::createScene(const osg::Vec3& gravity) {
	nextJointId = 0;
	PxSceneDesc sceneDesc(physics()->getTolerancesScale());
	
	sceneDesc.gravity = PxVec3(gravity[0], gravity[1], gravity[2]);
	sceneDesc.flags |= PxSceneFlag::eENABLE_CCD;

	if (!sceneDesc.cpuDispatcher) {
		sceneDesc.cpuDispatcher = PxDefaultCpuDispatcherCreate(1);
	}

	sceneDesc.filterShader = &TeetairFilterShader;

	scene = physics()->createScene(sceneDesc);
	if (!scene) {
		std::cout << "Unable to create Scene" << std::endl;
		return false;
	}

	scene->setVisualizationParameter(PxVisualizationParameter::eJOINT_LOCAL_FRAMES, 1.0f);
	scene->setVisualizationParameter(PxVisualizationParameter::eJOINT_LIMITS, 1.0f);
	scene->setSimulationEventCallback(this);

	return true;
}

bool PhysxScene::createMaterial(const string name, const PxReal staticFriction, const PxReal dynamicFriction, const PxReal restitution) {
	if (materials.count(name)) {
		std::cout << "There already is a material named " << name << std::endl;
		return false;
	}
	PxMaterial* material = physics()->createMaterial(staticFriction, dynamicFriction, restitution);
	if (!material) {
		std::cout << "Failed to create material" << std::endl;
		return false;
	}
	materials[name] = material;
	return true;
}

bool PhysxScene::createPlane(int id, const string materialName, const osg::Vec3 center, const osg::Vec3 normal, const PxU32 filterGroup, const PxU32 filterMask) {
	if (actors.count(id)) {
		std::cout << "There already is an actor with id " << id << std::endl;
		return false;
	}
	PxRigidStatic* actor = physics()->createRigidStatic(
			PxTransformFromPlaneEquation(PxPlane(PxVec3(center.x(), center.y(), center.z()), PxVec3(normal.x(), normal.y(), normal.z()))));
	if (!actor) {
		std::cout << "Failed to create actor " << id << std::endl;
		return false;
	}
	actors[id] = actor;
	
	PxShape* shape;
	if (!(shape = _addGeometryToActor(actor, &PxPlaneGeometry(), materialName, osg::Matrix::identity(), filterGroup, filterMask, false))) {
		return false;
	}

	return _finishActor(actor);
}

bool PhysxScene::createBox(int id, const osg::Vec3& dim, double density, const string materialName, const ForceResponsiveness forceResponsiveness, const PxU32 filterGroup, const PxU32 filterMask) {
	PxRigidDynamic* actor = _createActor(id, forceResponsiveness);
	if (!actor) {
		return false;
	}
	if (!_addGeometryToActor(actor, &PxBoxGeometry(PxVec3(dim[0] / 2.0, dim[1] / 2.0, dim[2] / 2.0)), materialName, osg::Matrix::identity(), filterGroup, filterMask, false)) {
		return false;
	}
	return _finishActor(actor, density);
}

bool PhysxScene::createSphere(int id, double radius, double density, const string materialName, const ForceResponsiveness forceResponsiveness, const PxU32 filterGroup, const PxU32 filterMask){
	PxRigidDynamic* actor = _createActor(id, forceResponsiveness);
	if (!actor) {
		return false;
	}
	if (!_addGeometryToActor(actor, &PxSphereGeometry(radius), materialName, osg::Matrix::identity(), filterGroup, filterMask, false)) {
		return false;
	}
	return _finishActor(actor, density);
}

bool PhysxScene::createConvexBox(int id, const osg::Vec3Array& vertices, double density, const string materialName, const ForceResponsiveness forceResponsiveness, const PxU32 filterGroup, const PxU32 filterMask) {
	PxRigidDynamic* actor = _createActor(id, forceResponsiveness);
	if (!actor) {
		return false;
	}
	
	if (!addConvexBoxToCompound(id, vertices, materialName, filterGroup, filterMask)) {
		return false;
	}

	return _finishActor(actor, density);
}

bool PhysxScene::createCompound(int id, const ForceResponsiveness forceResponsiveness){
	if (_createActor(id, forceResponsiveness))
		return true;
	return false;
}

bool PhysxScene::addBoxToCompound(int id, const osg::Vec3& dim, const string materialName, const osg::Matrix& localPose, const PxU32 filterGroup, const PxU32 filterMask, const bool triggerShape){
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return false;
	}
	return _addGeometryToActor(actors[id], &PxBoxGeometry(PxVec3(dim[0] / 2.0, dim[1] / 2.0, dim[2] / 2.0)), materialName, localPose, filterGroup, filterMask, triggerShape) != NULL;
}

bool PhysxScene::addConvexBoxToCompound(int id, const osg::Vec3Array& vertices, const string materialName, const PxU32 filterGroup, const PxU32 filterMask, const bool triggerShape) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return false;
	}

	PxVec3* pxVertices = new PxVec3[vertices.size()];
	for (int i = 0; i < vertices.size(); ++i) {
		pxVertices[i] = PxVec3(vertices[i].x(), vertices[i].y(), vertices[i].z());
	}

	PxConvexMeshDesc convexDesc;
	convexDesc.points.count = vertices.size();
	convexDesc.points.stride = sizeof(PxVec3);
	convexDesc.points.data = pxVertices;
	convexDesc.flags = PxConvexFlag::eCOMPUTE_CONVEX;

	PxCooking* cooking = PhysxManager::getCooking();
	PxDefaultMemoryOutputStream convexOutput;
	cooking->cookConvexMesh(convexDesc, convexOutput);
	PxDefaultMemoryInputData convexInput(convexOutput.getData(), convexOutput.getSize());
	PxConvexMesh* convexMesh = physics()->createConvexMesh(convexInput);

	if (!_addGeometryToActor(actors[id], &PxConvexMeshGeometry(convexMesh), materialName, osg::Matrix::identity(), filterGroup, filterMask, triggerShape)) {
		delete[] pxVertices;
		return false;
	}
	delete[] pxVertices;
	return true;
}

bool PhysxScene::finishCompound(int id, const double density){
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return false;
	}
	return _finishActor(actors[id], density);
}

PxRigidDynamic* PhysxScene::_createActor(int id, const ForceResponsiveness forceResponsiveness){
	if (actors.count(id)) {
		std::cout << "There already is an actor with id " << id << std::endl;
		return nullptr;
	}
	PxRigidDynamic* actor = physics()->createRigidDynamic(PxTransform::createIdentity());
	if (!actor) {
		std::cout << "Failed to create actor " << id << std::endl;
		return actor;
	}

	actors[id] = actor;

	setForceResponsiveness(id, forceResponsiveness);
	return actor;
}

PxShape* PhysxScene::_addGeometryToActor(PxRigidActor* actor, PxGeometry* geometry, const string materialName, const osg::Matrix& localPose, const PxU32 filterGroup, const PxU32 filterMask, const bool triggerShape) {
	PxShape* shape = actor->createShape(*geometry, *materials[materialName], PxTransform(toPxMat44(localPose)));
	if (!shape) {
		std::cout << "Failed to create shape for actor" << std::endl;
		return nullptr;
	}
	shape->setSimulationFilterData(PxFilterData(filterGroup, filterMask, 0, 0));
	if (triggerShape) {
		shape->setFlag(PxShapeFlag::eSIMULATION_SHAPE, false);
		shape->setFlag(PxShapeFlag::eTRIGGER_SHAPE, true);
	}
	return shape;
}

bool PhysxScene::_finishActor(PxRigidActor* actor, const double density) {
	if (actor->isRigidDynamic()) {
		PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
		PxRigidBodyExt::updateMassAndInertia(*dynamicActor, density);
	}
	scene->addActor(*actor);
	return true;
}

bool PhysxScene::createSphericalJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1) {
	PxRigidActor* actor0 = (actors.count(actorId0) ? actors[actorId0] : NULL);
	PxRigidActor* actor1 = (actors.count(actorId1) ? actors[actorId1] : NULL);

	if (actor0 == actor1) {
		std::cout << "I cannot create a joint between an object and itself!" << std::endl;
		return false;
	}
	PxSphericalJoint* joint = PxSphericalJointCreate(*physics(), actor0, PxTransform(toPxMat44(localPose0)), actor1, PxTransform(toPxMat44(localPose1)));
	joint->setConstraintFlag(PxConstraintFlag::eVISUALIZATION, true);
	joints[jointId] = joint;
	return true;
}

bool PhysxScene::setSphericalJointLimit(int id, PxReal yLimitAngle, PxReal zLimitAngle, PxReal limitContactDistance) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return false;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::eSPHERICAL)) {
		std::cout << "Joint " << id << " is not a spherical joint! He cannot have a spherical joint limit." << std::endl;
		return false;
	}
	PxSphericalJoint* sphericalJoint = static_cast<PxSphericalJoint*>(joint);
	sphericalJoint->setLimitCone(PxJointLimitCone(yLimitAngle,zLimitAngle, limitContactDistance));
	sphericalJoint->setSphericalJointFlag(PxSphericalJointFlag::eLIMIT_ENABLED, true);
	return true;
}

bool PhysxScene::setMaxAngularVelocity(int id, PxReal maxVelocity) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return false;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. He cannot have a maximum angular velocity." << std::endl;
		return false;
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	dynamicActor->setMaxAngularVelocity(maxVelocity);

	return true;
}

bool PhysxScene::createDistanceJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1) {
	PxRigidActor* actor0 = (actors.count(actorId0) ? actors[actorId0] : NULL);
	PxRigidActor* actor1 = (actors.count(actorId1) ? actors[actorId1] : NULL);

	if (actor0 == actor1) {
		std::cout << "I cannot create a joint between an object and itself!" << std::endl;
		return false;
	}
	PxDistanceJoint* joint = PxDistanceJointCreate(*physics(), actor0, PxTransform(toPxMat44(localPose0)), actor1, PxTransform(toPxMat44(localPose1)));
	joint->setConstraintFlag(PxConstraintFlag::eVISUALIZATION, true);
	joints[jointId] = joint;
	return true;
}

bool PhysxScene::setDistanceJointLimit(int id, PxReal minDistance, PxReal maxDistance, PxReal tolerance, PxReal stiffness) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return false;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::eDISTANCE)) {
		std::cout << "Joint " << id << " is not a distance joint! He cannot have a distance joint limit." << std::endl;
		return false;
	}
	PxDistanceJoint* distanceJoint = static_cast<PxDistanceJoint*>(joint);
	distanceJoint->setMinDistance(minDistance);
	distanceJoint->setDistanceJointFlag(PxDistanceJointFlag::eMIN_DISTANCE_ENABLED, minDistance >= 0);
	distanceJoint->setMaxDistance(maxDistance);
	distanceJoint->setDistanceJointFlag(PxDistanceJointFlag::eMAX_DISTANCE_ENABLED, maxDistance >= 0);
	distanceJoint->setTolerance(tolerance);
	distanceJoint->setStiffness(stiffness);
	return true;
}

bool PhysxScene::createFixedJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1) {
	PxRigidActor* actor0 = (actors.count(actorId0) ? actors[actorId0] : NULL);
	PxRigidActor* actor1 = (actors.count(actorId1) ? actors[actorId1] : NULL);

	if (actor0 == actor1) {
		std::cout << "I cannot create a joint between an object and itself!" << std::endl;
		return false;
	}
	PxFixedJoint* joint = PxFixedJointCreate(*physics(), actor0, PxTransform(toPxMat44(localPose0)), actor1, PxTransform(toPxMat44(localPose1)));
	joint->setConstraintFlag(PxConstraintFlag::eVISUALIZATION, true);
	joints[jointId] = joint;
	return true;
}

bool PhysxScene::createRevoluteJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1) {
	PxRigidActor* actor0 = (actors.count(actorId0) ? actors[actorId0] : NULL);
	PxRigidActor* actor1 = (actors.count(actorId1) ? actors[actorId1] : NULL);

	if (actor0 == actor1) {
		std::cout << "I cannot create a joint between an object and itself!" << std::endl;
		return false;
	}
	PxRevoluteJoint* joint = PxRevoluteJointCreate(*physics(), actor0, PxTransform(toPxMat44(localPose0)), actor1, PxTransform(toPxMat44(localPose1)));
	joint->setConstraintFlag(PxConstraintFlag::eVISUALIZATION, true);
	joints[jointId] = joint;
	return true;
}

bool PhysxScene::setRevoluteJointLimit(int id, PxReal lowerAngle, PxReal upperAngle, PxReal limitContactDistance) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return false;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::eREVOLUTE)) {
		std::cout << "Joint " << id << " is not a revolute joint! He cannot have a revolute joint limit." << std::endl;
		return false;
	}
	PxRevoluteJoint* revoluteJoint = static_cast<PxRevoluteJoint*>(joint);
	revoluteJoint->setLimit(PxJointAngularLimitPair(lowerAngle, upperAngle, limitContactDistance));
	revoluteJoint->setRevoluteJointFlag(PxRevoluteJointFlag::eLIMIT_ENABLED, true);
	return true;
}

bool PhysxScene::createPrismaticJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1) {
	PxRigidActor* actor0 = (actors.count(actorId0) ? actors[actorId0] : NULL);
	PxRigidActor* actor1 = (actors.count(actorId1) ? actors[actorId1] : NULL);

	if (actor0 == actor1) {
		std::cout << "I cannot create a joint between an object and itself!" << std::endl;
		return false;
	}
	PxPrismaticJoint* joint = PxPrismaticJointCreate(*physics(), actor0, PxTransform(toPxMat44(localPose0)), actor1, PxTransform(toPxMat44(localPose1)));
	joint->setConstraintFlag(PxConstraintFlag::eVISUALIZATION, true);
	joints[jointId] = joint;
	return true;
}
bool PhysxScene::setPrismaticJointLimit(int id, PxReal minDistance, PxReal maxDistance, PxReal contactDistance) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return false;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::ePRISMATIC)) {
		std::cout << "Joint " << id << " is not a prismatic joint! He cannot have a prismatic joint limit." << std::endl;
		return false;
	}
	PxPrismaticJoint* prismaticJoint = static_cast<PxPrismaticJoint*>(joint);
	prismaticJoint->setLimit(PxJointLinearLimitPair(physics()->getTolerancesScale(), minDistance, maxDistance, contactDistance));
	prismaticJoint->setPrismaticJointFlag(PxPrismaticJointFlag::eLIMIT_ENABLED, true);
	return true;
}

bool PhysxScene::createD6Joint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1) {
	PxRigidActor* actor0 = (actors.count(actorId0) ? actors[actorId0] : NULL);
	PxRigidActor* actor1 = (actors.count(actorId1) ? actors[actorId1] : NULL);

	if (actor0 == actor1) {
		std::cout << "I cannot create a joint between an object and itself!" << std::endl;
		return false;
	}
	PxD6Joint* joint = PxD6JointCreate(*physics(), actor0, PxTransform(toPxMat44(localPose0)), actor1, PxTransform(toPxMat44(localPose1)));
	joint->setConstraintFlag(PxConstraintFlag::eVISUALIZATION, true);
	joints[jointId] = joint;
	return true;
}
bool PhysxScene::setD6Motion(int id, PxD6Axis::Enum axis, PxD6Motion::Enum type, PxReal extent, PxReal contactDistance) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return false;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::eD6)) {
		std::cout << "Joint " << id << " is not a D6 joint! He cannot have a D6 joint limit." << std::endl;
		return false;
	}
	PxD6Joint* d6Joint = static_cast<PxD6Joint*>(joint);
	d6Joint->setMotion(axis, type);
	if (type == PxD6Motion::Enum::eLIMITED) {
		d6Joint->setLinearLimit(PxJointLinearLimit(physics()->getTolerancesScale(), extent, contactDistance));
	}
	return true;
}
bool PhysxScene::setD6Drive(int id, PxD6Drive::Enum axis, PxReal driveSpring, PxReal driveDamping, PxReal forceLimit, bool isAcceleration) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return false;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::eD6)) {
		std::cout << "Joint " << id << " is not a D6 joint! He cannot have a D6 drive." << std::endl;
		return false;
	}
	PxD6Joint* d6Joint = static_cast<PxD6Joint*>(joint);
	d6Joint->setDrive(axis, PxD6JointDrive(driveSpring, driveDamping, forceLimit, isAcceleration));
	return true;
}
void PhysxScene::d6Drive(int id, const osg::Matrix& targetPose, const osg::Vec3& linearVelocity, const osg::Vec3& angularVelocity) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return;
	}
	PxJoint* joint = joints[id];
	if (!(joint->getType() == PxJointType::eD6)) {
		std::cout << "Joint " << id << " is not a D6 joint but a " << joint->getConcreteTypeName() << "! He cannot drive." << std::endl;
		return;
	}
	PxD6Joint* d6Joint = static_cast<PxD6Joint*>(joint);
	deferredSceneChangesMutex.lock();
	osg::Matrix* pTargetPose = new osg::Matrix(targetPose);
	deferredSceneChanges.push(
		[d6Joint, pTargetPose, linearVelocity, angularVelocity]() {
		d6Joint->setDrivePosition(PxTransform(toPxMat44(*pTargetPose)));
		d6Joint->setDriveVelocity(PxVec3(linearVelocity.x(), linearVelocity.y(), linearVelocity.z()), PxVec3(angularVelocity.x(), angularVelocity.y(), angularVelocity.z()));
		PxRigidActor* actor0, *actor1;
		d6Joint->getActors(actor0, actor1);
		if (actor0->isRigidDynamic()) {
			static_cast<PxRigidDynamic*>(actor0)->wakeUp();
		}
		if (actor1->isRigidDynamic()) {
			static_cast<PxRigidDynamic*>(actor1)->wakeUp();
		}
		delete pTargetPose;
	});
	deferredSceneChangesMutex.unlock();
	
	return;
}

void PhysxScene::releaseJoint(int id) {
	if (!joints.count(id)) {
		std::cout << "I do not know a joint with id " << id << std::endl;
		return;
	}
	PxJoint* joint = joints[id];
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[joint]() {
		joint->release();
	});
	joints.erase(id);
	deferredSceneChangesMutex.unlock();

	return;
}

osg::Matrix PhysxScene::toOsgMatrix(const PxMat44 &matrix){
	float m[16];
	for (int i = 0; i < 16; ++i) m[i] = *(matrix.front() + i);
	return osg::Matrix(&m[0]);
}
PxMat44 PhysxScene::toPxMat44(const osg::Matrix &matrix){
	PxReal m[16];
	for (int i = 0; i < 16; ++i) m[i] = *(matrix.ptr() + i);
	return PxMat44(m);
}

osg::Matrix PhysxScene::getMatrix(int id) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return osg::Matrix();
	}
	PxRigidActor* actor = actors[id];

	return toOsgMatrix(PxMat44(actor->getGlobalPose()));
}

void PhysxScene::setMatrix(int id, const osg::Matrix& matrix) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	osg::Matrix* pMatrix = new osg::Matrix(matrix);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[actor, pMatrix]() {
		actor->setGlobalPose(PxTransform(toPxMat44(*pMatrix)));
		delete pMatrix;
	});
	deferredSceneChangesMutex.unlock();
	return;
}

void PhysxScene::addForce(int id, const osg::Vec3& force, PxForceMode::Enum mode, bool autowake) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	osg::Vec3* pForce = new osg::Vec3(force);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[id, actor, pForce, mode, autowake](){
		if (!actor->isRigidDynamic()) {
			std::cout << "The actor with id " << id << " is not dynamic. The force cannot be with him." << std::endl;
			return;
		}

		PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
		if (dynamicActor->getRigidDynamicFlags() & PxRigidDynamicFlag::eKINEMATIC) {
			std::cout << "The actor with id " << id << " is kinematic. He won't respond to forces." << std::endl;
			return;
		}
		dynamicActor->addForce(PxVec3((*pForce)[0], (*pForce)[1], (*pForce)[2]), mode, autowake);
		delete pForce;
	});
	deferredSceneChangesMutex.unlock();
	return;
}
void PhysxScene::clearForce(int id, PxForceMode::Enum mode) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];

	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. The force cannot be with him." << std::endl;
		return;
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	if (dynamicActor->getRigidDynamicFlags() & PxRigidDynamicFlag::eKINEMATIC) {
		std::cout << "The actor with id " << id << " is kinematic. He won't respond to forces." << std::endl;
		return;
	}

	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[dynamicActor, mode](){
		dynamicActor->clearForce(mode);
	});
	deferredSceneChangesMutex.unlock();
	return;
}

void PhysxScene::addTorque(int id, const osg::Vec3& torque, PxForceMode::Enum mode, bool autowake) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	osg::Vec3* pTorque = new osg::Vec3(torque);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[id, actor, pTorque, mode, autowake](){
		if (!actor->isRigidDynamic()) {
			std::cout << "The actor with id " << id << " is not dynamic. The force cannot be with him." << std::endl;
			return;
		}

		PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
		if (dynamicActor->getRigidDynamicFlags() & PxRigidDynamicFlag::eKINEMATIC) {
			std::cout << "The actor with id " << id << " is kinematic. He won't respond to forces." << std::endl;
			return;
		}
		dynamicActor->addTorque(PxVec3((*pTorque)[0], (*pTorque)[1], (*pTorque)[2]), mode, autowake);
		delete pTorque;
	});
	deferredSceneChangesMutex.unlock();
	return;
}

bool PhysxScene::setAngularDamping(int id, const PxReal angularDamping){
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return false;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. He can't have damping." << std::endl;
		return false;
	}
	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	dynamicActor->setAngularDamping(angularDamping);
	return true;
}
bool PhysxScene::setLinearDamping(int id, const PxReal linearDamping){
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return false;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. He can't have damping." << std::endl;
		return false;
	}
	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	dynamicActor->setLinearDamping(linearDamping);
	return true;
}

void PhysxScene::setLinearVelocity(int id, const osg::Vec3& linVel, bool autowake) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	osg::Vec3* pLinVel = new osg::Vec3(linVel);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[id, actor, pLinVel, autowake](){
		if (!actor->isRigidDynamic()) {
			std::cout << "The actor with id " << id << " is not dynamic. I cannot set a velocity." << std::endl;
			return;
		}

		PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
		if (dynamicActor->getRigidDynamicFlags() & PxRigidDynamicFlag::eKINEMATIC) {
			std::cout << "The actor with id " << id << " is kinematic. He won't respond to velocities." << std::endl;
			return;
		}
		dynamicActor->setLinearVelocity(PxVec3((*pLinVel)[0], (*pLinVel)[1], (*pLinVel)[2]), autowake);
		delete pLinVel;
	});
	deferredSceneChangesMutex.unlock();
	return;
}

osg::Vec3 PhysxScene::getLinearVelocity(int id) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return osg::Vec3();
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. I cannot read a velocity." << std::endl;
		return osg::Vec3();
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	PxVec3 linVel = dynamicActor->getLinearVelocity();
	return osg::Vec3(linVel.x, linVel.y, linVel.z);
}

void PhysxScene::setAngularVelocity(int id, const osg::Vec3& angVel, bool autowake) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	osg::Vec3* pAngVel = new osg::Vec3(angVel);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[id, actor, pAngVel, autowake](){
		if (!actor->isRigidDynamic()) {
			std::cout << "The actor with id " << id << " is not dynamic. I cannot set a velocity." << std::endl;
			return;
		}

		PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
		if (dynamicActor->getRigidDynamicFlags() & PxRigidDynamicFlag::eKINEMATIC) {
			std::cout << "The actor with id " << id << " is kinematic. He won't respond to velocities." << std::endl;
			return;
		}
		dynamicActor->setAngularVelocity(PxVec3((*pAngVel)[0], (*pAngVel)[1], (*pAngVel)[2]), autowake);
		delete pAngVel;
	});
	deferredSceneChangesMutex.unlock();
	return;
}

osg::Vec3 PhysxScene::getAngularVelocity(int id) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return osg::Vec3();
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. I cannot read a velocity." << std::endl;
		return osg::Vec3();
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	PxVec3 angVel = dynamicActor->getAngularVelocity();
	return osg::Vec3(angVel.x, angVel.y, angVel.z);
}

void PhysxScene::setSimulateCollisions(int id, bool simulate) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[actor, simulate](){
		PxShape** shapes = new PxShape*[actor->getNbShapes()];
		actor->getShapes(shapes, actor->getNbShapes());
		for (unsigned int i = 0; i < actor->getNbShapes(); ++i) {
			if (shapes[i]->getFlags() & PxShapeFlag::eTRIGGER_SHAPE)
				continue;
			shapes[i]->setFlag(PxShapeFlag::eSIMULATION_SHAPE, simulate);
		}
		delete[] shapes;
	});
	deferredSceneChangesMutex.unlock();
}

void PhysxScene::setSleepThreshold(int id, PxReal threshold) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. I cannot respond to forces." << std::endl;
		return;
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[dynamicActor, threshold](){
		dynamicActor->setSleepThreshold(threshold);
	});
	deferredSceneChangesMutex.unlock();
}

void PhysxScene::setForceResponsiveness(int id, ForceResponsiveness forceResponsiveness) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. I cannot respond to forces." << std::endl;
		return;
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[dynamicActor, forceResponsiveness](){
		switch (forceResponsiveness)
		{
		case NO_GRAVITY:
			dynamicActor->setRigidDynamicFlag(PxRigidDynamicFlag::eKINEMATIC, false);
			dynamicActor->setActorFlag(PxActorFlag::eDISABLE_GRAVITY, true);
			break;
		case KINEMATIC:
			dynamicActor->setRigidDynamicFlag(PxRigidDynamicFlag::eKINEMATIC, true);
			break;
		case ALL:
			dynamicActor->setRigidDynamicFlag(PxRigidDynamicFlag::eKINEMATIC, false);
			dynamicActor->setActorFlag(PxActorFlag::eDISABLE_GRAVITY, false);
			break;
		default:
			break;
		}
	});
	deferredSceneChangesMutex.unlock();
}

void PhysxScene::setEnableCCD(int id, bool enable) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidDynamic()) {
		std::cout << "The actor with id " << id << " is not dynamic. I cannot use CCD." << std::endl;
		return;
	}

	PxRigidDynamic* dynamicActor = static_cast<PxRigidDynamic*>(actor);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[dynamicActor, enable](){
		dynamicActor->setRigidBodyFlag(PxRigidBodyFlag::eENABLE_CCD, enable);
	});
	deferredSceneChangesMutex.unlock();
}

void PhysxScene::setSendSleepNotifies(int id, bool enable) {
	if (!actors.count(id)) {
		std::cout << "I do not know an actor with id " << id << std::endl;
		return;
	}
	PxRigidActor* actor = actors[id];
	if (!actor->isRigidBody()) {
		std::cout << "The actor with id " << id << " is not a rigid body. I cannot notify of sleep events." << std::endl;
		return;
	}

	PxRigidBody* rigidActor = static_cast<PxRigidBody*>(actor);
	deferredSceneChangesMutex.lock();
	deferredSceneChanges.push(
		[rigidActor, enable](){
		rigidActor->setActorFlag(PxActorFlag::eSEND_SLEEP_NOTIFIES, enable);
	});
	deferredSceneChangesMutex.unlock();
}

bool PhysxScene::areActors(int actorId0, int actorId1, PxRigidActor* const contactActors[2]) {
	if (!actors.count(actorId0) || !actors.count(actorId1)) {
		return false;
	}
	return (contactActors[0] == actors[actorId0] && contactActors[1] == actors[actorId1])
		|| (contactActors[0] == actors[actorId1] && contactActors[1] == actors[actorId0]);
}

bool PhysxScene::containsActor(int actorId, PxRigidActor* const contactActors[2]) {
	if (!actors.count(actorId)) {
		return false;
	}
	return (contactActors[0] == actors[actorId] || contactActors[1] == actors[actorId]);
}

bool PhysxScene::isActor(int actorId, PxActor* const actor) {
	if (!actors.count(actorId)) {
		return false;
	}
	return (actor == actors[actorId]);
}

bool PhysxScene::areFilterGroups(int filterGroup0, int filterGroup1, PxShape* const contactShapes[2]) {
	return (contactShapes[0]->getSimulationFilterData().word0 == filterGroup0 && contactShapes[1]->getSimulationFilterData().word0 == filterGroup1)
		|| (contactShapes[0]->getSimulationFilterData().word0 == filterGroup1 && contactShapes[1]->getSimulationFilterData().word0 == filterGroup0);
}

bool PhysxScene::containsFilterGroup(int filterGroup, PxShape* const contactShapes[2]) {
	return (contactShapes[0]->getSimulationFilterData().word0 == filterGroup || contactShapes[1]->getSimulationFilterData().word0 == filterGroup);
}


void PhysxScene::simulate(double step) {
	// Process deferred scene changes
	deferredSceneChangesMutex.lock();
	while (!deferredSceneChanges.empty()) {
		deferredSceneChanges.front()();
		deferredSceneChanges.pop();
	}
	deferredSceneChangesMutex.unlock();
	// Simulate physx
	scene->simulate(step);
	// wait for calculation
	scene->fetchResults(true);
}

void PhysxScene::clearPhysxScene() {
	nextJointId = 0;
	for (JointMap::iterator i = joints.begin(); i != joints.end(); ++i) {
		i->second->release();
	}
	joints.clear();
	for (ActorMap::iterator i = actors.begin(); i != actors.end(); ++i) {
		i->second->release();
	}
	actors.clear();
	for (MaterialMap::iterator i = materials.begin(); i != materials.end(); ++i) {
		i->second->release();
	}
	materials.clear();
	deferredSceneChangesMutex.lock();
	std::queue<std::function<void()>>().swap(deferredSceneChanges);
	deferredSceneChangesMutex.unlock();
}

PhysxScene::~PhysxScene() {
	clearPhysxScene();
	if (scene)
		scene->release();
}


PxPhysics* PhysxScene::physics() {
	return PhysxManager::getPhysics();
}