#pragma once

#include <unordered_map>

#include "PhysxManager.h"

#include <osg/Referenced>
#include <osg/ref_ptr>
#include <osg/Vec3>
#include <osg/Geometry>
#include <osg/Plane>
#include <osg/Matrix>

#include <functional>
#include <queue>

#include <QMutex>

class PhysxScene : public osg::Referenced, public physx::PxSimulationEventCallback {
public:
	/// <summary>
	/// Creates the scene.
	/// </summary>
	/// <param name="gravity">The gravity vector.</param>
	/// <returns>true if the scene was created, false if the scene could not be created.</returns>
	bool createScene(const osg::Vec3& gravity);

	/// <summary>
	/// Creates a material.
	/// </summary>
	/// <param name="name">The name.</param>
	/// <param name="staticFriction">The static friction.</param>
	/// <param name="dynamicFriction">The dynamic friction.</param>
	/// <param name="restitution">The restitution.</param>
	/// <returns>true if the material was created, false if the material could not be created or if the name was already taken.</returns>
	bool createMaterial(const string name, const PxReal staticFriction, const PxReal dynamicFriction, const PxReal restitution);
	
	/// <summary>
	/// The behaviour of an actor concerning forces.
	/// </summary>
	enum ForceResponsiveness
	{
		/// <summary>
		/// All forces apply to the actor
		/// </summary>
		ALL,
		/// <summary>
		/// The actor won't notice gravity
		/// </summary>
		NO_GRAVITY,
		/// <summary>
		/// The actor is kinematic and does not respond to forces at all
		/// </summary>
		KINEMATIC
	};

	/// <summary>
	/// Creates a plane.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="center">The center point.</param>
	/// <param name="normal">The normal vector.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <returns>true if the plane was created, false if the actor could not be created or the id is already taken.</returns>
	bool createPlane(int id, const string materialName, const osg::Vec3 center, const osg::Vec3 normal, const PxU32 filterGroup = 0, const PxU32 filterMask = 0);

	/// <summary>
	/// Creates a box.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="dim">The dimension of the box (width, length, height).</param>
	/// <param name="density">The density.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="forceResponsiveness">The force responsiveness.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <returns>true if the box was created, false if the actor could not be created or the id is already taken.</returns>
	bool createBox(int id, const osg::Vec3& dim, double density, const string materialName, const ForceResponsiveness forceResponsiveness = ALL, const PxU32 filterGroup = 0, const PxU32 filterMask = 0);

	/// <summary>
	/// Creates a sphere.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="radius">The radius.</param>
	/// <param name="density">The density.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="forceResponsiveness">The force responsiveness.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <returns>true if the sphere was created, false if the actor could not be created or the id is already taken.</returns>
	bool createSphere(int id, double radius, double density, const string materialName, const ForceResponsiveness forceResponsiveness = ALL, const PxU32 filterGroup = 0, const PxU32 filterMask = 0);

	/// <summary>
	/// Creates a convex box.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="vertices">The vertices of the convex object.</param>
	/// <param name="density">The density.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="forceResponsiveness">The force responsiveness.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <returns>true if the convex box was created, false if the actor could not be created or the id is already taken.</returns>
	bool createConvexBox(int id, const osg::Vec3Array& vertices, double density, const string materialName, const ForceResponsiveness forceResponsiveness, const PxU32 filterGroup = 0, const PxU32 filterMask = 0);

	/// <summary>
	/// Creates a compound actor, consisting of several shapes.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="forceResponsiveness">The force responsiveness.</param>
	/// <returns>true if the compound was created, false if the actor could not be created or the id is already taken.</returns>
	bool createCompound(int id, const ForceResponsiveness forceResponsiveness);
	/// <summary>
	/// Adds a box to a previously created compound.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="dim">The dim.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="localPose">The local pose.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <param name="triggerShape">if set to <c>true</c>, the shape will not play a role in the simulation, but only produce trigger events.</param>
	/// <returns>true if the box was added to the compound, false if the compound does not exist or the box could not be added.</returns>
	bool addBoxToCompound(int id, const osg::Vec3& dim, const string materialName, const osg::Matrix& localPose, const PxU32 filterGroup = 0, const PxU32 filterMask = 0, const bool triggerShape = false);
	/// <summary>
	/// Adds a convex box to a previously created compound.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="vertices">The vertices of the convex object.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <param name="triggerShape">if set to <c>true</c>, the shape will not play a role in the simulation, but only produce trigger events.</param>
	/// <returns>true if the box was added to the compound, false if the compound does not exist or the box could not be added.</returns>
	bool addConvexBoxToCompound(int id, const osg::Vec3Array& vertices, const string materialName, const PxU32 filterGroup = 0, const PxU32 filterMask = 0, const bool triggerShape = false);
	/// <summary>
	/// Finishes the compound and adds the actor to the scene.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="density">The density.</param>
	/// <returns>true if the compound was finished, false if the compound does not exist or the actor could not be added to the scene</returns>
	bool finishCompound(int id, const double density = 1.0);

private:
	/// <summary>
	/// Creates an actor.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="forceResponsiveness">The force responsiveness.</param>
	/// <returns>A pointer to the created RigidDynamic actor or a nullptr if the actor could not be created or the id is already taken.</returns>
	PxRigidDynamic* _createActor(int id, const ForceResponsiveness forceResponsiveness);
	/// <summary>
	/// Adds the geometry to the actor.
	/// </summary>
	/// <param name="actor">The actor.</param>
	/// <param name="geometry">The geometry.</param>
	/// <param name="materialName">Name of the material.</param>
	/// <param name="localPose">The local pose of the geometry relative to the actor.</param>
	/// <param name="filterGroup">The filter group.</param>
	/// <param name="filterMask">The filter mask.</param>
	/// <param name="triggerShape">if set to <c>true</c>, the shape will not play a role in the simulation, but only produce trigger events.</param>
	/// <returns>The created Shape or a nullptr if the shape could not be created.</returns>
	PxShape* _addGeometryToActor(PxRigidActor* actor, PxGeometry* geometry, const string materialName, const osg::Matrix& localPose, const PxU32 filterGroup, const PxU32 filterMask, const bool triggerShape);
	/// <summary>
	/// Finishes the actor and adds the actor to the scene.
	/// </summary>
	/// <param name="actor">The actor.</param>
	/// <param name="density">The density.</param>
	/// <returns>true</returns>
	bool _finishActor(PxRigidActor* actor, const double density = 0.0);
	
public:
	/// <summary>
	/// Gets the next joint identifier.
	/// </summary>
	/// <returns>the identifier to use for the next joint creation.</returns>
	int getNextJointId() { return nextJointId++; }

	/// <summary>
	/// Creates a spherical joint.
	/// </summary>
	/// <param name="jointId">The joint identifier.</param>
	/// <param name="actorId0">The first actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose0">The local pose of the first actor.</param>
	/// <param name="actorId1">The second actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose1">The local pose of the second actor.</param>
	/// <returns>true if the joint was created, false if the two actors are the same.</returns>
	bool createSphericalJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1);
	/// <summary>
	/// Sets the spherical joint limit.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="yLimitAngle">The y limit angle.</param>
	/// <param name="zLimitAngle">The z limit angle.</param>
	/// <param name="limitContactDistance">The limit contact distance.</param>
	/// <returns>true if the limit was set, false if the joint does not exist or is not a spherical joint.</returns>
	bool setSphericalJointLimit(int id, PxReal yLimitAngle, PxReal zLimitAngle, PxReal limitContactDistance);

	/// <summary>
	/// Sets the maximum angular velocity of an actor.
	/// </summary>
	/// <param name="id">The identifier of the actor.</param>
	/// <param name="maxVelocity">The maximum angular velocity.</param>
	/// <returns>true if the maximum was set, false if the actor does not exist or is not a dynmaic actor.</returns>
	bool setMaxAngularVelocity(int id, PxReal maxVelocity);

	/// <summary>
	/// Creates a distance joint.
	/// </summary>
	/// <param name="jointId">The joint identifier.</param>
	/// <param name="actorId0">The first actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose0">The local pose of the first actor.</param>
	/// <param name="actorId1">The second actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose1">The local pose of the second actor.</param>
	/// <returns>true if the joint was created, false if the two actors are the same.</returns>
	bool createDistanceJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1);
	/// <summary>
	/// Sets the distance joint limits.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="minDistance">The minimum distance.</param>
	/// <param name="maxDistance">The maximum distance.</param>
	/// <param name="tolerance">The tolerance of the joint.</param>
	/// <param name="stiffness">The stiffness of the joint.</param>
	/// <returns>true if the limits were set, false if the joint does not exist or is not a distance joint.</returns>
	bool setDistanceJointLimit(int id, PxReal minDistance, PxReal maxDistance, PxReal tolerance, PxReal stiffness);

	/// <summary>
	/// Creates a fixed joint.
	/// </summary>
	/// <param name="jointId">The joint identifier.</param>
	/// <param name="actorId0">The first actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose0">The local pose of the first actor.</param>
	/// <param name="actorId1">The second actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose1">The local pose of the second actor.</param>
	/// <returns>true if the joint was created, false if the two actors are the same.</returns>
	bool createFixedJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1);

	/// <summary>
	/// Creates a revolute joint.
	/// </summary>
	/// <param name="jointId">The joint identifier.</param>
	/// <param name="actorId0">The first actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose0">The local pose of the first actor.</param>
	/// <param name="actorId1">The second actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose1">The local pose of the second actor.</param>
	/// <returns>true if the joint was created, false if the two actors are the same.</returns>
	bool createRevoluteJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1);
	/// <summary>
	/// Sets the revolute joint limits.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="lowerAngle">The lower angle.</param>
	/// <param name="upperAngle">The upper angle.</param>
	/// <param name="contactDistance">The contact distance limit.</param>
	/// <returns>
	/// true if the limits were set, false if the joint does not exist or is not a revolute joint.
	/// </returns>
	bool setRevoluteJointLimit(int id, PxReal lowerAngle, PxReal upperAngle, PxReal contactDistance);

	/// <summary>
	/// Creates a prismatic joint.
	/// </summary>
	/// <param name="jointId">The joint identifier.</param>
	/// <param name="actorId0">The first actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose0">The local pose of the first actor.</param>
	/// <param name="actorId1">The second actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose1">The local pose of the second actor.</param>
	/// <returns>true if the joint was created, false if the two actors are the same.</returns>
	bool createPrismaticJoint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1);
	/// <summary>
	/// Sets the prismatic joint limits.
	/// </summary>
	/// <param name="id">The identifier.</param>
	/// <param name="lowerAngle">The lower angle.</param>
	/// <param name="upperAngle">The upper angle.</param>
	/// <param name="contactDistance">The contact distance limit.</param>
	/// <returns>
	/// true if the limits were set, false if the joint does not exist or is not a revolute joint.
	/// </returns>
	bool setPrismaticJointLimit(int id, PxReal minDistance, PxReal maxDistance, PxReal contactDistance);

	/// <summary>
	/// Creates a D6 joint.
	/// </summary>
	/// <param name="jointId">The joint identifier.</param>
	/// <param name="actorId0">The first actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose0">The local pose of the first actor.</param>
	/// <param name="actorId1">The second actor's id or an invalid id to joint to the world.</param>
	/// <param name="localPose1">The local pose of the second actor.</param>
	/// <returns>true if the joint was created, false if the two actors are the same.</returns>
	bool createD6Joint(int jointId, int actorId0, const osg::Matrix& localPose0, int actorId1, const osg::Matrix& localPose1);
	/// <summary>
	/// Sets a D6 joint axis motion.
	/// </summary>
	/// <param name="id">The joint identifier.</param>
	/// <param name="axis">The axis.</param>
	/// <param name="type">The motion type (locked/free/limited).</param>
	/// <param name="extent">The extent value if motion is limited.</param>
	/// <param name="contactDistance">The contact distance limit is limited.</param>
	/// <returns>
	/// true if the limits were set, false if the joint does not exist or is not a d6 joint.
	/// </returns>
	bool setD6Motion(int id, PxD6Axis::Enum axis, PxD6Motion::Enum type, PxReal extent = 0, PxReal contactDistance = 0);
	/// <summary>
	/// Sets the d6 drive for an axis.
	/// </summary>
	/// <param name="id">The joint identifier.</param>
	/// <param name="axis">The axis.</param>
	/// <param name="driveSpring">The drive's spring.</param>
	/// <param name="driveDamping">The drive's damping.</param>
	/// <param name="forceLimit">The spring's force limit.</param>
	/// <param name="isAcceleration">if set to <c>true</c> the spring will apply to acceleration instead of force.</param>
	/// <returns>
	/// true if the drive was set, false if the joint does not exist or is not a d6 joint.
	/// </returns>
	bool setD6Drive(int id, PxD6Drive::Enum axis, PxReal driveSpring, PxReal driveDamping, PxReal forceLimit, bool isAcceleration = false);
	/// <summary>
	/// Drive a D6 joint.
	/// </summary>
	/// <param name="id">The joint identifier.</param>
	/// <param name="targetPose">The target pose.</param>
	/// <param name="linearVelocity">The linear velocity.</param>
	/// <param name="angularVelocity">The angular velocity.</param>
	/// <returns>
	/// true if the parameters were set, false if the joint does not exist or is not a d6 joint.
	/// </returns>
	void d6Drive(int id, const osg::Matrix& targetPose, const osg::Vec3& linearVelocity, const osg::Vec3& angularVelocity);

	/// <summary>
	/// Releases the joint.
	/// </summary>
	/// <param name="id">The identifier.</param>
	void releaseJoint(int id);

public:
	/// <summary>
	/// Converts a physX matrix to a osg matrix.
	/// </summary>
	/// <param name="matrix">The physX matrix.</param>
	/// <returns>The osg matrix</returns>
	static osg::Matrix toOsgMatrix(const PxMat44 &matrix);
	/// <summary>
	/// Converts a osg matrix to a osg matrix.
	/// </summary>
	/// <param name="matrix">The osg matrix.</param>
	/// <returns>The physX matrix</returns>
	static PxMat44 toPxMat44(const osg::Matrix &matrix);
	
	/// <summary>
	/// Gets the matrix of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <returns>The current pose matrix of the actor or an empty matrix if the actor does not exist.</returns>
	osg::Matrix getMatrix(int id);
	/// <summary>
	/// Sets the current pose of the actor to the matrix.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="matrix">The pose matrix.</param>
	void setMatrix(int id, const osg::Matrix& matrix);
	
	/// <summary>
	/// Adds the force to the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="force">The force.</param>
	/// <param name="mode">The force mode.</param>
	/// <param name="autowake">if set to <c>true</c>, wake the actor if he's sleeping.</param>
	void addForce(int id, const osg::Vec3& force, PxForceMode::Enum mode = PxForceMode::eFORCE, bool autowake = true);
	/// <summary>
	/// Clears the forces staged for the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="mode">The force mode to clear.</param>
	void clearForce(int id, PxForceMode::Enum mode = PxForceMode::eFORCE);
	/// <summary>
	/// Adds the torque to the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="torque">The torque vector (axis, length is amount).</param>
	/// <param name="mode">The force mode.</param>
	/// <param name="autowake">if set to <c>true</c>, wake the actor if he's sleeping.</param>
	void addTorque(int id, const osg::Vec3& torque, PxForceMode::Enum mode = PxForceMode::eFORCE, bool autowake = true);

	/// <summary>
	/// Sets the angular damping of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="angularDamping">The angular damping.</param>
	/// <returns>true if the damping was set, false if the actor does not exist or is not dynamic</returns>
	bool setAngularDamping(int id, const PxReal angularDamping);
	/// <summary>
	/// Sets the linear damping of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="linearDamping">The linear damping.</param>
	/// <returns>true if the damping was set, false if the actor does not exist or is not dynamic</returns>
	bool setLinearDamping(int id, const PxReal linearDamping);

	/// <summary>
	/// Sets the linear velocity of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="linVel">The linear velocity.</param>
	/// <param name="autowake">if set to <c>true</c>, wake the actor if he's sleeping.</param>
	void setLinearVelocity(int id, const osg::Vec3& linVel, bool autowake = true);
	/// <summary>
	/// Gets the linear velocity of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <returns>The linear velocity of the actor or an empty vector if the actor does not exist or is not dynamic.</returns>
	osg::Vec3 getLinearVelocity(int id);
	/// <summary>
	/// Sets the angular velocity of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="angVel">The angular velocity.</param>
	/// <param name="autowake">if set to <c>true</c>, wake the actor if he's sleeping.</param>
	void setAngularVelocity(int id, const osg::Vec3& angVel, bool autowake = true);
	/// <summary>
	/// Gets the angular velocity of the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <returns>The angular velocity of the actor or an empty vector if the actor does not exist or is not dynamic.</returns>
	osg::Vec3 getAngularVelocity(int id);

	/// <summary>
	/// (De-)activates collision tests for the actor.
	/// The parameter will be set on all shapes of the actor except for trigger shapes.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="simulate">if set to <c>true</c> the actor will be considered during collision tests.</param>
	void setSimulateCollisions(int id, bool simulate);

	/// <summary>
	/// Sets the sleep threshold.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="threshold">The sleep threshold.</param>
	void setSleepThreshold(int id, PxReal threshold);

	/// <summary>
	/// Sets the force responsiveness.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="forceResponsiveness">The force responsiveness.</param>
	void setForceResponsiveness(int id, ForceResponsiveness forceResponsiveness);

	/// <summary>
	/// Sets if continuous collision detection is activated for the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="enable">if set to <c>true</c>. the actor won't go through other objects at high velocities.</param>
	void setEnableCCD(int id, bool enable);

	/// <summary>
	/// Sets whether onSleep notification is active for the actor.
	/// </summary>
	/// <param name="id">The actor's identifier.</param>
	/// <param name="enable">if set to <c>true</c>. the actor will notify the onSleep function when it sleeps.</param>
	void setSendSleepNotifies(int id, bool enable);

	/// <summary>
	/// Simulates the scene for the specified step count.
	/// </summary>
	/// <param name="step">The elapsed time to simulate.</param>
	void simulate(double step);

public:
	virtual void onConstraintBreak(PxConstraintInfo* constraints, PxU32 count){};
	virtual void onWake(PxActor** actors, PxU32 count){};
	virtual void onSleep(PxActor** actors, PxU32 count){};
	virtual void onContact(const PxContactPairHeader& pairHeader, const PxContactPair* pairs, PxU32 nbPairs){};
	virtual void onTrigger(PxTriggerPair* pairs, PxU32 count){};

	/// <summary>
	/// Checks if the two collided actors correspond to the two given actor ids.
	/// </summary>
	/// <param name="actorId0">The first actor's id.</param>
	/// <param name="actorId1">The second actor's id.</param>
	/// <param name="contactActors">The contact actors.</param>
	/// <returns>true if the actors with the given ids match the collided actors</returns>
	virtual bool areActors(int actorId0, int actorId1, PxRigidActor* const contactActors[2]);
	/// <summary>
	/// Checks if one of the two collided actors correspond to the given actor id.
	/// </summary>
	/// <param name="actorId">The actor's id.</param>
	/// <param name="contactActors">The contact actors.</param>
	/// <returns>true if the actor with the given id matchs one of the collided actors</returns>
	virtual bool containsActor(int actorId, PxRigidActor* const contactActors[2]);
	// <summary>
	/// Checks if the the actor corresponds to the given actor id.
	/// </summary>
	/// <param name="actorId">The actor's id.</param>
	/// <param name="actor">The actor.</param>
	/// <returns>true if the given id matches the actor.</returns>
	virtual bool isActor(int actorId, PxActor* const actor);
	/// <summary>
	/// Checks if the two filter groups correspond to the contactShapes.
	/// </summary>
	/// <param name="filterGroup1">The first filterGroup.</param>
	/// <param name="filterGroup2">The second filterGroup.</param>
	/// <param name="contactShapes">The contact shapes.</param>
	/// <returns>true if the contact shapes match the given filterGroups.</returns>
	virtual bool areFilterGroups(int filterGroup1, int filterGroup2, PxShape* const contactShapes[2]);
	/// <summary>
	/// Checks if the one of the contactShapes corresponds to the filterGroup.
	/// </summary>
	/// <param name="filterGroup">The filterGroup.</param>
	/// <param name="contactShapes">The contact shapes.</param>
	/// <returns>true if one of the contact shapes matchs the given filterGroup.</returns>
	virtual bool containsFilterGroup(int filterGroup, PxShape* const contactShapes[2]);

	/// <summary>
	/// Clears the physx scene.
	/// Releases and deletes all actors, joints and materials.
	/// Deletes the unprocessed deferredSceneChanges.
	/// </summary>
	void clearPhysxScene();

protected:
	virtual ~PhysxScene();

friend class TeetairApplication;
private:
		typedef unordered_map<int, PxRigidActor*> ActorMap;
		ActorMap actors;
private:
	static PxPhysics* physics();
	
	typedef unordered_map<string, PxMaterial*> MaterialMap;
	MaterialMap materials;
	typedef unordered_map<int, PxJoint*> JointMap;
	JointMap joints;
	PxScene* scene;

	std::queue<std::function<void()>> deferredSceneChanges;
	mutable QMutex deferredSceneChangesMutex;
	int nextJointId;
};

typedef PhysxScene PX;